from unittest import mock

import ehrax as rx
import pytest
import tables as tb
from ehrax import CodeMap, CodingScheme, CodingSchemesManager, FrozenDict1N
from ehrax.example_schemes.icd_ccs_integration import (
    CCSSchemeSelection,
    ICDSchemeSelection,
    OutcomeSelection,
    setup_ccs_schemes,
    setup_icd_schemes,
    setup_outcomes,
    setup_standard_icd_ccs,
)


class TestFlatScheme:
    @pytest.fixture(
        scope="class",
        params=[
            dict(name="one", codes=["1"], desc={"1": "one"}),
            dict(name="zero", codes=[], desc=dict()),
            dict(name="100", codes=list(f"code_{i}" for i in range(100))),
        ],
    )
    def primitive_flat_scheme_kwarg(self, request):
        if "desc" in request.param:
            desc = request.param["desc"]
        elif len(request.param["codes"]) > 0:
            desc = dict(zip(request.param["codes"], request.param["codes"]))
        else:
            desc = dict()
        return dict(name=request.param["name"], codes=tuple(sorted(request.param["codes"])), desc=rx.FrozenDict11(desc))

    @pytest.fixture(scope="class")
    def primitive_flat_scheme(self, primitive_flat_scheme_kwarg) -> rx.CodingScheme:
        return rx.CodingScheme(**primitive_flat_scheme_kwarg)

    @pytest.fixture(scope="class")
    def scheme_manager(self, primitive_flat_scheme: rx.CodingScheme) -> rx.CodingSchemesManager:
        return rx.CodingSchemesManager().add_scheme(primitive_flat_scheme)

    def test_from_name(self, primitive_flat_scheme: rx.CodingScheme, scheme_manager: rx.CodingSchemesManager):
        assert scheme_manager.scheme[primitive_flat_scheme.name].equals(primitive_flat_scheme)

        with pytest.raises(KeyError):
            # Unregistered scheme
            _ = scheme_manager.scheme["42"]

    @pytest.mark.parametrize("codes", [("A", "B", "C", "C"), ("A", "B", "C", "B"), ("A", "A", "A", "A")])
    def test_codes_uniqueness(self, codes):
        with pytest.raises(AssertionError) as excinfo:
            _ = rx.CodingScheme(name="test", codes=tuple(sorted(codes)), desc=rx.FrozenDict11({c: c for c in codes}))
            assert "should be unique" in str(excinfo.value)

    def test_register_scheme(self, primitive_flat_scheme):
        """
        Test the register_scheme method.

        This method tests two scenarios:
        1. It tests that the register_scheme method works by registering a scheme and then
           asserting that the registered scheme can be retrieved using its name.
        2. It tests that the register_scheme method logs a warning when trying to
           register a scheme that is already registered with the same name and content.
        """
        # First, test that the register_scheme method works.
        manager = rx.CodingSchemesManager().add_scheme(primitive_flat_scheme)
        assert manager.scheme[primitive_flat_scheme.name].equals(primitive_flat_scheme)

        # Second, test that the register_scheme method raises an error when
        # the scheme is already registered.
        with mock.patch("logging.warning") as mocker:
            manager.add_scheme(primitive_flat_scheme)
            mocker.assert_called_once()

    def test_scheme_equality(self, primitive_flat_scheme):
        """
        Test the equality of schemes.

        This test asserts that a scheme equal to its deepcopy.
        It then mutates the description and index of one of the schemes and asserts that the two
        schemes are not equal.
        """
        assert primitive_flat_scheme.equals(primitive_flat_scheme)

        if len(primitive_flat_scheme) > 0:
            desc_mutated = rx.FrozenDict11({code: f"{desc} muted" for code, desc in primitive_flat_scheme.desc.items()})
            mutated_scheme = rx.CodingScheme(
                name=primitive_flat_scheme.name, codes=primitive_flat_scheme.codes, desc=desc_mutated
            )
            assert not primitive_flat_scheme.equals(mutated_scheme)

    @pytest.fixture(
        params=[
            (("icd10cm",), ()),
            (("icd9cm",), ()),
            (("icd10pcs",), ()),
            (("icd9pcs",), ()),
            (("icd9cm",), ("dx_ccs",)),
            (("icd9pcs",), ("pr_ccs",)),
            (("icd9cm",), ("dx_flat_ccs",)),
            (("icd9pcs",), ("pr_flat_ccs",)),
        ],
        ids=lambda x: "_".join(sum(x, ())),
        scope="class",
    )
    def scheme_selection(self, request) -> tuple[ICDSchemeSelection, CCSSchemeSelection]:
        kwargs = lambda p: {k: True for k in p}
        return ICDSchemeSelection(**kwargs(request.param[0])), CCSSchemeSelection(**kwargs(request.param[1]))

    @pytest.fixture(
        params=[
            (("icd10cm", "icd9cm"), ()),
            (("icd9cm", "icd10cm"), ()),
            (("icd10pcs", "icd9pcs"), ()),
            (("icd10pcs", "icd9pcs"), ()),
            (("icd9cm",), ("dx_ccs",)),
            (("icd9pcs",), ("pr_ccs",)),
            (("icd9cm",), ("dx_flat_ccs",)),
            (("icd9pcs",), ("pr_flat_ccs",)),
        ],
        scope="class",
        ids=lambda x: "_".join(sum(x, ())),
    )
    def scheme_pair_selection(self, request):
        kwargs = lambda p: {k: True for k in p}
        return ICDSchemeSelection(**kwargs(request.param[0])), CCSSchemeSelection(**kwargs(request.param[1]))

    @pytest.fixture(
        params=["icd9cm_v1", "icd9cm_v2_groups", "icd9cm_v3_groups", "dx_flat_ccs_mlhc_groups", "dx_flat_ccs_v1"],
        scope="class",
    )
    def outcome_selection(self, request):
        return OutcomeSelection(**{request.param: True})

    @pytest.fixture(scope="class")
    def outcome_selection_name(self, outcome_selection: OutcomeSelection) -> str:
        (name,) = outcome_selection.flag_set
        return name

    @pytest.fixture(scope="class")
    def selection_names(self, scheme_selection: tuple[ICDSchemeSelection, CCSSchemeSelection]) -> tuple[str, ...]:
        return tuple(scheme_selection[0].flag_set) + tuple(scheme_selection[1].flag_set)

    @pytest.fixture(scope="class")
    def icd_ccs_scheme_manager(
        self, scheme_selection: tuple[ICDSchemeSelection, CCSSchemeSelection]
    ) -> rx.CodingSchemesManager:
        icd_selection, ccs_selection = scheme_selection
        return setup_ccs_schemes(setup_icd_schemes(icd_selection), ccs_selection)

    @pytest.fixture(scope="class")
    def icd_ccs_outcome_manager_prerequisite(self, outcome_selection: OutcomeSelection) -> rx.CodingSchemesManager:
        return setup_ccs_schemes(
            setup_icd_schemes(ICDSchemeSelection(icd9cm=True)), CCSSchemeSelection(dx_flat_ccs=True)
        )

    @pytest.fixture(scope="class")
    def icd_ccs_map_manager(
        self, scheme_pair_selection: tuple[ICDSchemeSelection, CCSSchemeSelection]
    ) -> rx.CodingSchemesManager:
        icd_selection, ccs_selection = scheme_pair_selection
        return setup_standard_icd_ccs(
            icd_selection=icd_selection, ccs_selection=ccs_selection, outcome_selection=OutcomeSelection()
        )

    @pytest.fixture(scope="class")
    def icd_ccs_outcome_manager(
        self, icd_ccs_outcome_manager_prerequisite: rx.CodingSchemesManager, outcome_selection: OutcomeSelection
    ) -> rx.CodingSchemesManager:
        return setup_outcomes(icd_ccs_outcome_manager_prerequisite, outcome_selection)

    def test_icd_ccs_schemes(self, icd_ccs_scheme_manager, selection_names):
        assert len(icd_ccs_scheme_manager.schemes) == len(selection_names)
        assert len(icd_ccs_scheme_manager.scheme) == len(selection_names)
        for name in selection_names:
            assert name in icd_ccs_scheme_manager.scheme
            assert icd_ccs_scheme_manager.scheme[name].name == name
            assert isinstance(icd_ccs_scheme_manager.scheme[name], rx.CodingScheme)

    def test_icd_ccs_outcomes(self, icd_ccs_outcome_manager, outcome_selection_name):
        assert len(icd_ccs_outcome_manager.outcomes) == 1
        assert len(icd_ccs_outcome_manager.outcome_data) == 1

        assert outcome_selection_name in icd_ccs_outcome_manager.outcome_data
        assert icd_ccs_outcome_manager.outcome_data[outcome_selection_name].name == outcome_selection_name
        assert isinstance(icd_ccs_outcome_manager.outcome_data[outcome_selection_name], rx.FilterOutcomeMapData)

    def test_icd_ccs_maps(
        self,
        icd_ccs_map_manager: rx.CodingSchemesManager,
        scheme_pair_selection: tuple[ICDSchemeSelection, CCSSchemeSelection],
    ):
        icd_selection, ccs_selection = scheme_pair_selection
        assert len(icd_selection.flag_set) + len(ccs_selection.flag_set) == 2
        assert len(icd_ccs_map_manager.maps) == 2
        assert len(icd_ccs_map_manager.map) == 2 + 2  # the two identity maps
        (a, b) = icd_selection.flag_set + ccs_selection.flag_set
        s_a = icd_ccs_map_manager.scheme[a]
        s_b = icd_ccs_map_manager.scheme[b]
        assert (a, b) in icd_ccs_map_manager.map
        assert (b, a) in icd_ccs_map_manager.map
        m1 = icd_ccs_map_manager.map[(a, b)]
        m2 = icd_ccs_map_manager.map[(b, a)]
        for m in (m1, m2):
            assert set(m.data.keys()) == set(m.domain)
            assert frozenset().union(*tuple(m.data.values())) == frozenset(m.range)
        assert m1.source_name == a
        assert m1.target_name == b
        assert m2.source_name == b
        assert m2.target_name == a
        assert m1.source_name == a
        assert m1.target_name == b
        assert m2.source_name == b
        assert m2.target_name == a
        assert m1.support_ratio(s_a) > 0.7
        assert m2.support_ratio(s_b) > 0.7
        assert m1.range_ratio(s_b) > 0.15
        assert m2.range_ratio(s_a) > 0.15

    def test_primitive_scheme_serialization(self, primitive_flat_scheme: rx.CodingScheme, tmpdir: str):
        path = f"{tmpdir}/coding_scheme.h5"
        with tb.open_file(path, "w") as f:
            primitive_flat_scheme.to_hdf_group(f.create_group("/", "scheme_data"))

        with tb.open_file(path, "r") as f:
            reloaded = rx.CodingScheme.from_hdf_group(f.root.scheme_data)

        assert primitive_flat_scheme.equals(reloaded)

    def test_icd_ccs_scheme_serialization(self, icd_ccs_scheme_manager: rx.CodingSchemesManager, tmpdir: str):
        path = f"{tmpdir}/coding_schemes.h5"
        with tb.open_file(path, "w") as f:
            icd_ccs_scheme_manager.to_hdf_group(f.create_group("/", "context_view"))

        with tb.open_file(path, "r") as f:
            reloaded = rx.CodingSchemesManager.from_hdf_group(f.root.context_view)

        assert icd_ccs_scheme_manager.equals(reloaded)

    def test_icd_ccs_outcome_serialization(self, icd_ccs_outcome_manager: rx.CodingSchemesManager, tmpdir: str):
        path = f"{tmpdir}/coding_schemes.h5"
        with tb.open_file(path, "w") as f:
            icd_ccs_outcome_manager.to_hdf_group(f.create_group("/", "context_view"))

        with tb.open_file(path, "r") as f:
            reloaded = rx.CodingSchemesManager.from_hdf_group(f.root.context_view)

        assert icd_ccs_outcome_manager.equals(reloaded)

    def test_icd_ccs_map_serialization(
        self, icd_ccs_map_manager: rx.CodingSchemesManager, hf5_group_writer: tb.Group, tmpdir: str
    ):
        path = f"{tmpdir}/coding_schemes.h5"
        with tb.open_file(path, "w") as f:
            icd_ccs_map_manager.to_hdf_group(f.create_group("/", "context_view"))

        with tb.open_file(path, "r") as f:
            reloaded = rx.CodingSchemesManager.from_hdf_group(f.root.context_view)

        assert icd_ccs_map_manager.equals(reloaded)

    @pytest.mark.parametrize(
        "name, codes, desc",
        [
            ("problematic_codes", [1], {"1": "one"}),
            ("problematic_desc", ["1"], {1: "one"}),
            ("problematic_desc", ["1"], {"1": 5}),
        ],
    )
    def test_type_error(self, name: str, codes: list[str], desc: dict[str, str]):
        """
        Test for type error handling in the FlatScheme constructor.

        This test adds a problematic scheme to test the error handling of the constructor.
        The code and description types should be strings, not integers. The test expects an AssertionError
        or KeyError to be raised.
        """
        with pytest.raises((AssertionError, KeyError)):
            rx.CodingScheme(name=name, codes=tuple(sorted(codes)), desc=rx.FrozenDict11(desc))

    @pytest.mark.parametrize(
        "name, codes, desc",
        [
            ("problematic_desc", ["1", "3"], {"1": "one"}),
            ("problematic_desc", ["3"], {"3": "three", "1": "one"}),
            ("duplicate_codes", ["1", "2", "2"], {"1": "one", "2": "two"}),
        ],
    )
    def test_sizes(self, name: str, codes: list[str], desc: dict[str, str]):
        """
        Test the consistency between scheme components, in their size, and mapping correctness.

        This method adds a problematic scheme to test error handling.
        The codes, description, and index collections should all have the same sizes,
        codes should be unique, and mapping should be correct and 1-to-1.
        FlatScheme constructor raises either an AssertionError or KeyError when provided with invalid input.
        """
        with pytest.raises((AssertionError, KeyError)):
            rx.CodingScheme(name=name, codes=tuple(sorted(codes)), desc=rx.FrozenDict11(desc))

    def test_index2code(self, primitive_flat_scheme):
        """
        Test the index to code mapping in the coding scheme.

        It tests if the index to code mapping is correct and respects the codes order.
        """

        assert all(c == primitive_flat_scheme.codes[i] for i, c in primitive_flat_scheme.index2code.items())

    def test_index2desc(self, primitive_flat_scheme):
        """
        Test the mapping of index to description in the coding scheme.

        Iterates over the scheme_kwargs and creates a FlatScheme object using each set of keyword arguments.
        Then, it checks if the description for each code in the scheme matches the description obtained from the index.
        """

        assert all(
            desc == primitive_flat_scheme.desc[primitive_flat_scheme.codes[i]]
            for i, desc in primitive_flat_scheme.index2desc.items()
        )

    def test_search_regex(self):
        """
        Test the search_regex method of the FlatScheme class.

        This method tests the search_regex method of the FlatScheme class by creating a FlatScheme object
        with a specific coding scheme configuration and performing various search operations using the
        search_regex method. It asserts the expected results for each search operation.

        """
        # Arrange
        scheme = rx.CodingScheme(
            name="simple_searchable", codes=("1", "3"), desc=rx.FrozenDict11({"1": "one", "3": "pancreatic cAnCeR"})
        )
        # Act & Assert
        assert scheme.search_regex("cancer") == {"3"}
        assert scheme.search_regex("one") == {"1"}
        assert scheme.search_regex("pancreatic") == {"3"}
        assert scheme.search_regex("cAnCeR") == {"3"}
        assert scheme.search_regex("x") == set()

    # def test_mapper_to(self):
    #     self.fail()

    # def test_codeset2vec(self):
    #     self.fail()

    # def test_empty_vector(self):
    #     self.fail()

    # def test_supported_targets(self):
    #     self.fail()

    def test_as_dataframe(self, primitive_flat_scheme):
        """
        Test the `as_dataframe` method of the FlatScheme class.

        This method tests whether the `as_dataframe` method returns a DataFrame with the expected structure and values.
        It checks if the index of the DataFrame matches the index values of the scheme and if the columns of the
        DataFrame are 'code' and 'desc'. It also verifies if the scheme object is equal to a new FlatScheme object
        created with the same configuration, codes, descriptions, and index.
        """
        df = primitive_flat_scheme.as_dataframe()
        assert set(df.index) == set(primitive_flat_scheme.index.values())
        assert set(df.columns) == {"code", "desc"}
        codes = df.code.tolist()
        desc = df.set_index("code")["desc"].to_dict()
        assert primitive_flat_scheme.equals(
            rx.CodingScheme(name=primitive_flat_scheme.name, codes=tuple(sorted(codes)), desc=rx.FrozenDict11(desc))
        )


class TestReducedCodeMapN1:
    @pytest.fixture(scope="class")
    def codes_n1(self):
        return {
            "A1": ["B0", "B1", "B2"],
            "A2": ["B3", "B5", "B7"],
            "A3": ["B6"],
            "A4": ["B4", "B8"],
        }  # permute: 0, 1, 2, 3, 5, 7, 6, 4, 8

    @pytest.fixture(scope="class")
    def aggregation(self):
        return rx.FrozenDict11({"A1": "w_sum", "A2": "w_sum", "A3": "w_sum", "A4": "w_sum"})

    @pytest.fixture(scope="class")
    def source_code_scheme(self, codes_n1) -> rx.CodingScheme:
        codes = tuple(sorted(b for bs in codes_n1.values() for b in bs))
        desc = rx.FrozenDict11({b: b for b in codes})
        return rx.CodingScheme(name="source", codes=codes, desc=desc)

    @pytest.fixture(scope="class")
    def target_code_scheme(self, codes_n1) -> rx.CodingScheme:
        desc = rx.FrozenDict11({k: k for k in codes_n1.keys()})
        return rx.CodingScheme(name="target", codes=tuple(codes_n1.keys()), desc=desc)

    @pytest.fixture(scope="class")
    def mapping_data(self, codes_n1) -> rx.FrozenDict1N:
        return rx.FrozenDict1N({b: {a} for a, bs in codes_n1.items() for b in bs})

    @pytest.fixture(scope="class")
    def reduced_code_map(
        self, source_code_scheme, target_code_scheme, mapping_data, aggregation
    ) -> rx.ReducedCodeMapN1:
        return rx.ReducedCodeMapN1.from_data(
            source_code_scheme.name, target_code_scheme.name, mapping_data, aggregation
        )

    @pytest.fixture(scope="class")
    def scheme_manager(self, source_code_scheme, target_code_scheme, reduced_code_map):
        return (
            rx.CodingSchemesManager()
            .add_scheme(source_code_scheme)
            .add_scheme(target_code_scheme)
            .add_map(reduced_code_map)
        )

    @pytest.fixture(scope="class")
    def source_index(
        self, reduced_code_map: rx.ReducedCodeMapN1, scheme_manager: rx.CodingSchemesManager
    ) -> dict[str, int]:
        return scheme_manager.scheme[reduced_code_map.source_name].index

    def test_reduced_groups(self, reduced_code_map: rx.ReducedCodeMapN1):
        assert set(reduced_code_map.reduced_groups.data.keys()) == {"A1", "A2", "A3", "A4"}

    def test_groups_aggregation(self, reduced_code_map: rx.ReducedCodeMapN1):
        assert reduced_code_map.groups_aggregation == ("w_sum",) * 4

    def test_groups_size(self, reduced_code_map: rx.ReducedCodeMapN1, source_index: dict[str, int]):
        assert reduced_code_map.groups_size(source_index) == (3, 3, 1, 2)

    def test_groups_split(self, reduced_code_map: rx.ReducedCodeMapN1, source_index: dict[str, int]):
        assert reduced_code_map.groups_split(source_index) == (3, 6, 7, 9)

    def test_groups_permute(self, reduced_code_map: rx.ReducedCodeMapN1, source_index: dict[str, int]):
        assert reduced_code_map.groups_permute(source_index) == (0, 1, 2, 3, 5, 7, 6, 4, 8)


class TestCodingSchemesManager:
    @pytest.fixture(scope="class")
    def scheme1(self) -> CodingScheme:
        return CodingScheme(name="s", codes=("A", "B"))

    @pytest.fixture(scope="class")
    def scheme2(self) -> CodingScheme:
        return CodingScheme(name="t", codes=("C", "D"))

    @pytest.fixture(scope="class")  # to test chaining.
    def scheme3(self) -> CodingScheme:
        return CodingScheme(name="q", codes=("E", "F"))

    @pytest.fixture(scope="class")
    def map_a(self, scheme1: CodingScheme, scheme2: CodingScheme) -> CodeMap:
        return CodeMap(source_name=scheme1.name, target_name=scheme2.name, data=FrozenDict1N({"A": {"D"}}))

    @pytest.fixture(scope="class")  # to test chaining
    def map_b(self, scheme2: CodingScheme, scheme3: CodingScheme) -> CodeMap:
        return CodeMap(source_name=scheme2.name, target_name=scheme3.name, data=FrozenDict1N({"D": {"F"}}))

    @pytest.fixture(scope="class")
    def scheme_x(self) -> CodingScheme:  # to test match map
        return CodingScheme(name="x", codes=tuple(sorted(("xO", "hY", " zN"))))

    @pytest.fixture(scope="class")
    def scheme_y(self) -> CodingScheme:  # to test match map
        return CodingScheme(name="y", codes=tuple(sorted((" Xo", "Hy ", "ZN"))))

    @pytest.fixture(scope="class")
    def scheme_z(self) -> CodingScheme:  # to test match map
        return CodingScheme(name="z", codes=(" Xoo", "iHy ", "iiii"))

    @pytest.fixture(scope="class")
    def manager1(self, scheme1: CodingScheme) -> CodingSchemesManager:
        return CodingSchemesManager().add_scheme(scheme1)

    @pytest.fixture(scope="class")
    def manager2(self, scheme2: CodingScheme) -> CodingSchemesManager:
        return CodingSchemesManager().add_scheme(scheme2)

    @pytest.fixture(scope="class")
    def manager3(self, map_a: CodeMap) -> CodingSchemesManager:
        return CodingSchemesManager().add_map(map_a)

    @pytest.fixture(scope="class")
    def manager_union(
        self, manager1: CodingSchemesManager, manager2: CodingSchemesManager, manager3: CodingSchemesManager
    ) -> CodingSchemesManager:
        return manager1 + manager2 + manager3

    @pytest.fixture(scope="class")  # to test chaining
    def manager_all(
        self, manager_union: CodingSchemesManager, scheme3: CodingScheme, map_b: CodeMap
    ) -> CodingSchemesManager:
        return manager_union.add_scheme(scheme3).add_map(map_b)

    @pytest.fixture(scope="class")  # tot test chaining
    def manager_with_chained_map(self, manager_all: CodingSchemesManager):
        return manager_all.add_chained_map("s", "t", "q")

    @pytest.fixture(scope="class")
    def manager4matchmap(
        self, scheme_x: CodingScheme, scheme_y: CodingScheme, scheme_z: CodingScheme
    ) -> CodingSchemesManager:
        return CodingSchemesManager().add_scheme(scheme_x).add_scheme(scheme_y).add_scheme(scheme_z)

    @pytest.fixture(scope="class")
    def manager_with_match_map(self, manager4matchmap: CodingSchemesManager):
        return manager4matchmap.add_match_map("x", "y")

    def test_dims0(self):
        m = CodingSchemesManager()
        assert len(m.scheme) == 0
        assert len(m.map) == 0
        assert len(m.outcome_data) == 0
        assert len(m.identity_maps) == 0
        assert m.equals(CodingSchemesManager())

    def test_dims1(self, manager1: CodingSchemesManager, manager2: CodingSchemesManager):
        assert len(manager1.scheme) == len(manager2.scheme) == 1
        assert len(manager1.map) == len(manager2.scheme) == 1  # now it contains an identity map (automatically made)
        assert len(manager1.identity_maps) == len(manager2.identity_maps) == 1
        assert manager1.equals(manager1)
        assert not manager1.equals(manager2)
        assert not manager1.equals(CodingSchemesManager())

    def test_dims2(self, manager3: CodingSchemesManager):
        assert len(manager3.scheme) == 0
        assert len(manager3.map) == 1
        assert len(manager3.identity_maps) == 0
        assert not manager3.equals(CodingSchemesManager())
        assert manager3.equals(manager3)

    def test_dims3(
        self,
        manager1: CodingSchemesManager,
        manager2: CodingSchemesManager,
        manager3: CodingSchemesManager,
        manager_union: CodingSchemesManager,
    ):
        assert (manager1 + manager2).equals(manager2 + manager1)
        assert not (manager1 + manager2 + manager3).equals(manager1 + manager3)
        assert (manager1 + manager2 + manager3).equals(manager3 + manager2 + manager1 + CodingSchemesManager())
        assert (manager1 + manager1 + manager1).equals(manager1)
        assert len(manager_union.schemes) == 2
        assert len(manager_union.identity_maps) == 2
        assert len(manager_union.map) == 3

    def test_chained_map(self, manager_all: CodingSchemesManager, manager_with_chained_map: CodingSchemesManager):
        assert ("s", "t") in manager_all.map
        assert ("s", "t") in manager_with_chained_map.map
        assert ("s", "q") not in manager_all.map
        assert ("s", "q") in manager_with_chained_map.map
        s = manager_with_chained_map.scheme["s"]
        q = manager_with_chained_map.scheme["q"]
        s_q = manager_with_chained_map.map[("s", "q")]
        assert set(s.codes).issuperset(s_q.data.keys())
        assert s_q.map_codeset(s.codes).issubset(q.codes)

    def test_manager_with_match_map(self, manager_with_match_map: CodingSchemesManager):
        scheme_x = manager_with_match_map.scheme["x"]
        scheme_y = manager_with_match_map.scheme["y"]
        m_xy = manager_with_match_map.map[("x", "y")]
        m_yx = manager_with_match_map.map[("y", "x")]
        assert set(scheme_x.codes) == set(m_xy.data.keys())
        assert set(scheme_y.codes) == set(m_yx.data.keys())
        assert set(scheme_x.codes) == frozenset().union(*list(m_yx.data.values()))
        assert set(scheme_y.codes) == frozenset().union(*list(m_xy.data.values()))
        assert set(m_xy.map_codeset(scheme_x.codes)) == set(scheme_y.codes)
        assert set(m_yx.map_codeset(scheme_y.codes)) == set(scheme_x.codes)
        assert m_yx.map_codeset(m_xy.map_codeset(scheme_x.codes)) == set(scheme_x.codes)
        assert m_xy.map_codeset(m_yx.map_codeset(scheme_y.codes)) == set(scheme_y.codes)

    def test_invalid_match_map(self, manager4matchmap: CodingSchemesManager):
        with pytest.raises(AssertionError):
            _ = manager4matchmap.add_match_map("x", "z")



import random
from abc import ABCMeta, abstractmethod
from collections.abc import Callable, Hashable, Generator
from typing import Any, cast

import equinox as eqx
import numpy as np
import pandas as pd

from ._literals import SplitLiteral
from .coding_scheme import CodeMap, CodingSchemesManager, NumericScheme
from .dataset import (
    AbstractTransformation,
    AdmissionIntervalEventsTableColumns,
    AdmissionIntervalRatesTableColumns,
    Dataset,
    Report,
)
from .transformations import DatasetTransformation
from .tvx_concepts import (
    Admission,
    CodesVector,
    InpatientInput,
    InpatientInterventions,
    InpatientObservables,
    LeadingObservableExtractor,
    Patient,
    StaticInfo,
)
from .tvx_ehr import (
    CodedValueProcessor,
    CodedValueScaler,
    IQROutlierRemoverConfig,
    ScalerConfig,
    SegmentedTVxEHR,
    TVxEHR,
    TVxReport,
    TVxReportAttributes,
)
from .utils import dataframe_log


def dataset_surgery(getter: Callable[[TVxEHR], Any], dataset: TVxEHR, replacement: Any) -> TVxEHR:
    return eqx.tree_at(getter, dataset, replacement)


class TrainableTransformation(AbstractTransformation, metaclass=ABCMeta):
    # dependencies: ClassVar[tuple[type[DatasetTransformation], ...]] = (RandomSplits, setIndex, SetCodeIntegerIndices)

    @staticmethod
    def get_admission_ids(tvx_ehr: TVxEHR) -> list[str]:
        c_subject_id = tvx_ehr.dataset.config.columns.static.subject_id
        c_admission_id = tvx_ehr.dataset.config.columns.admissions.admission_id
        admissions = tvx_ehr.dataset.tables.admissions[[c_subject_id]]
        assert c_admission_id in admissions.index.names, f"Column {c_admission_id} not found in admissions table index."
        assert tvx_ehr.splits is not None, "TVxEHR.splits is None"
        training_subject_ids = tvx_ehr.splits[0]
        return admissions.loc[admissions.loc[:, c_subject_id].isin(training_subject_ids)].index.unique().tolist()


class TVxTransformation(AbstractTransformation[TVxEHR, TVxReport]):
    @classmethod
    @abstractmethod
    def apply(
        cls, tvx_ehr: TVxEHR, schemes_context: CodingSchemesManager, report: TVxReport
    ) -> tuple[TVxEHR, TVxReport]: ...


class SampleSubjects(TVxTransformation):
    @classmethod
    def apply(
        cls, tvx_ehr: TVxEHR, schemes_context: CodingSchemesManager, report: TVxReport
    ) -> tuple[TVxEHR, TVxReport]:
        static = tvx_ehr.dataset.tables.static
        # assert index name is subject_id
        c_subject_id = tvx_ehr.dataset.config.columns.static.subject_id
        assert c_subject_id in static.index.names, f"Index name must be {c_subject_id}"
        config = tvx_ehr.config.sample
        if config is None:
            return cls.skip(tvx_ehr, report, reason="config.sample is not configured.")

        rng = random.Random(config.seed)
        subjects = static.index.unique().tolist()
        rng.shuffle(subjects)
        subjects = subjects[config.offset : config.offset + config.n_subjects]
        n1 = len(static)
        static = static.loc[subjects]
        n2 = len(static)
        report = report.add(
            table="static",
            column="index",
            before=n1,
            after=n2,
            value_type="count",
            transformation=cls,
            operation="sample",
        )
        dataset = eqx.tree_at(lambda x: x.tables.static, tvx_ehr.dataset, static)
        r = cast(Report, report)
        dataset, r = DatasetTransformation.synchronize_subjects(dataset, r)
        tvx_ehr = eqx.tree_at(lambda x: x.dataset, tvx_ehr, dataset)
        return tvx_ehr, cast(TVxReport, r)


class RandomSplits(TVxTransformation):
    @classmethod
    def apply(
        cls, tvx_ehr: TVxEHR, schemes_context: CodingSchemesManager, report: TVxReport
    ) -> tuple[TVxEHR, TVxReport]:
        config = tvx_ehr.config.splits

        if config is None:
            return cls.skip(tvx_ehr, report, "config.split is not configured.")

        splits = tvx_ehr.dataset.random_splits(
            splits=config.split_quantiles,
            random_seed=config.seed,
            balance=config.balance,
            discount_first_admission=config.discount_first_admission,
        )

        report = report.add(
            table="static",
            column=None,
            value_type="splits",
            transformation=cls,
            operation="TVxEHR.splits<-TVxEHR.dataset.random_splits(TVxEHR.config.splits)",
            before=(len(tvx_ehr.dataset.tables.static),),
            after=tuple(len(x) for x in splits),
        )
        tvx_ehr = eqx.tree_at(lambda x: x.splits, tvx_ehr, splits, is_leaf=lambda x: x is None)
        return tvx_ehr, report


class TrainingSplitGroups(TVxTransformation):
    @classmethod
    def sync_dataset(cls, dataset: Dataset, subject_ids: tuple[str, ...]) -> Dataset:
        static = dataset.tables.static
        c_subject_id = dataset.config.columns.static.subject_id
        assert c_subject_id in static.index.names, f"Index name must be {c_subject_id}"
        static = static.loc[list(subject_ids)]
        dataset = eqx.tree_at(lambda x: x.tables.static, dataset, static)
        return DatasetTransformation.synchronize_subjects(dataset, Report())[0]

    @classmethod
    def apply(
        cls, tvx_ehr: TVxEHR, schemes_context: CodingSchemesManager, report: TVxReport
    ) -> tuple[TVxEHR, TVxReport]:
        raise NotImplementedError("Use TrainingSplitGroups.split_dataset()")

    @classmethod
    def subset(cls, tvx_ehr: TVxEHR, group: tuple[str, ...]) -> TVxEHR:
        dataset = cls.sync_dataset(tvx_ehr.dataset, group)
        assert tvx_ehr.subjects is not None, "tvx_ehr.subjects is None"
        subjects = {subject_id: subject for subject_id, subject in tvx_ehr.subjects.items() if subject_id in group}
        tvx_ehr = eqx.tree_at(lambda x: x.dataset, tvx_ehr, dataset)
        tvx_ehr = eqx.tree_at(lambda x: x.subjects, tvx_ehr, subjects)
        tvx_ehr = eqx.tree_at(lambda x: x.splits, tvx_ehr, None, is_leaf=lambda x: x is None)
        return tvx_ehr

    @classmethod
    def split_dataset(
        cls, tvx_ehr: TVxEHR, n_groups: int, seed: int = 0, split_balance: SplitLiteral | None = None
    ) -> tuple[TVxEHR, ...]:
        assert tvx_ehr.config.splits is not None and tvx_ehr.splits is not None, "splits is None"
        training_split = tvx_ehr.splits[0]
        dataset = cls.sync_dataset(tvx_ehr.dataset, training_split)
        split_quantiles = np.linspace(0, 1, n_groups + 1)[1:-1]
        groups = dataset.random_splits(
            splits=split_quantiles.tolist(),
            random_seed=seed,
            balance=split_balance or tvx_ehr.config.splits.balance,  # type: ignore
            discount_first_admission=tvx_ehr.config.splits.discount_first_admission,
        )
        return tuple(cls.subset(tvx_ehr, tuple(group)) for group in groups)


class ZScoreScaler(CodedValueScaler):
    mean: pd.Series
    std: pd.Series

    def __init__(
        self,
        config: ScalerConfig,
        mean: pd.Series = pd.Series(),
        std: pd.Series = pd.Series(),
        table_name: str | None = None,
        code_column: str | None = None,
        value_column: str | None = None,
    ):
        super().__init__(config=config, table_name=table_name, code_column=code_column, value_column=value_column)
        self.mean = mean
        self.std = std

    @property
    def original_dtype(self) -> np.dtype:
        return self.mean.dtype  # type: ignore

    def __call__(self, dataset: Dataset) -> Dataset:
        table = self.table_getter(dataset)

        mean = table.loc[:, self.code_column].map(self.mean)
        std = table.loc[:, self.code_column].map(self.std)
        table.loc[:, self.value_column] = (table.loc[:, self.value_column] - mean) / std
        if self.config.use_float16:
            table = table.astype({self.value_column: np.float16})

        return eqx.tree_at(lambda x: self.table_getter(dataset), dataset, table)

    def unscale(self, array: np.ndarray) -> np.ndarray:
        array = array.astype(self.original_dtype)
        index = np.arange(array.shape[-1])
        return array * self.std.loc[index].values + self.mean.loc[index].values

    def unscale_code(self, array: np.ndarray, code_index: Hashable) -> np.ndarray:
        array = array.astype(self.original_dtype)
        return array * self.std.loc[code_index] + self.mean.loc[code_index]

    def _extract_stats(self, df: pd.DataFrame, c_code: str, c_value: str) -> dict[str, pd.Series]:
        stat = df.groupby(c_code)[[c_value]].apply(
            lambda x: pd.Series({"mu": x[c_value].mean(), "sigma": x[c_value].std()})
        )
        return dict(mean=stat.loc[:, "mu"], std=stat.loc[:, "sigma"])


class MaxScaler(CodedValueScaler):
    max_val: pd.Series

    def __init__(
        self,
        config: ScalerConfig,
        max_val: pd.Series = pd.Series(),
        table_name: str | None = None,
        code_column: str | None = None,
        value_column: str | None = None,
    ):
        super().__init__(config=config, table_name=table_name, code_column=code_column, value_column=value_column)
        self.max_val = max_val

    @property
    def original_dtype(self) -> np.dtype:
        return self.max_val.dtype  # type: ignore

    def __call__(self, dataset: Dataset) -> Dataset:
        df = self.table_getter(dataset).copy()

        max_val = df.loc[:, self.code_column].map(self.max_val)
        df.loc[:, self.value_column] = df.loc[:, self.value_column] / max_val
        if self.config.use_float16:
            df = df.astype({self.value_column: np.float16})
        return eqx.tree_at(self.table_getter, dataset, df)

    def unscale(self, array: np.ndarray) -> np.ndarray:
        array = array.astype(self.original_dtype)
        if array.shape[-1] == len(self.max_val):
            index = np.arange(array.shape[-1])
            return array * self.max_val.loc[index].values
        index = np.array(self.max_val.index.values)
        array = array.copy()
        if array.ndim == 1:
            array[index] *= self.max_val.values
        else:
            array[:, index] *= self.max_val.values
        return array

    def unscale_code(self, array: np.ndarray, code_index: Hashable) -> np.ndarray:
        array = array.astype(self.original_dtype)
        return array * self.max_val.loc[code_index]

    def _extract_stats(self, df: pd.DataFrame, c_code: str, c_value: str) -> dict[str, pd.Series]:
        stat = df.groupby(c_code)[[c_value]].apply(lambda x: pd.Series({"max": x[c_value].max()}))
        return dict(max_val=stat.loc[:, "max"])


class AdaptiveScaler(CodedValueScaler):
    max_val: pd.Series
    min_val: pd.Series
    mean: pd.Series
    std: pd.Series

    def __init__(
        self,
        config: ScalerConfig,
        max_val: pd.Series = pd.Series(),
        min_val: pd.Series = pd.Series(),
        mean: pd.Series = pd.Series(),
        std: pd.Series = pd.Series(),
        table_name: str | None = None,
        code_column: str | None = None,
        value_column: str | None = None,
    ):
        super().__init__(config=config, table_name=table_name, code_column=code_column, value_column=value_column)
        self.max_val = max_val
        self.min_val = min_val
        self.mean = mean
        self.std = std

    @property
    def original_dtype(self) -> np.dtype:
        return self.max_val.dtype  # type: ignore

    def __call__(self, dataset: Dataset) -> Dataset:
        df = self.table_getter(dataset).copy()

        min_val = df.loc[:, self.code_column].map(self.min_val)
        max_val = df.loc[:, self.code_column].map(self.max_val)
        mean = df.loc[:, self.code_column].map(self.mean)
        std = df.loc[:, self.code_column].map(self.std)

        minmax_scaled = (df.loc[:, self.value_column] - min_val) / max_val
        z_scaled = (df.loc[:, self.value_column] - mean) / std

        df.loc[:, self.value_column] = np.where(min_val >= 0.0, minmax_scaled, z_scaled)
        if self.config.use_float16:
            df = df.astype({self.value_column: np.float16})
        return eqx.tree_at(self.table_getter, dataset, df)

    def unscale(self, array: np.ndarray) -> np.ndarray:
        array = array.astype(self.original_dtype)
        index = np.arange(array.shape[-1])
        mu = self.mean.loc[index].values
        sigma = self.std.loc[index].values
        min_val = self.min_val.loc[index].values
        max_val = self.max_val.loc[index].values
        z_unscaled = array * sigma + mu
        minmax_unscaled = array * max_val + min_val
        return np.where(min_val >= 0.0, minmax_unscaled, z_unscaled)

    def unscale_code(self, array: np.ndarray, code_index: Hashable) -> np.ndarray:
        array = array.astype(self.original_dtype)
        mu = self.mean.loc[code_index]  # type: ignore
        sigma = self.std.loc[code_index]  # type: ignore
        min_val = self.min_val.loc[code_index]  # type: ignore
        max_val = self.max_val.loc[code_index]  # type: ignore
        z_unscaled = array * sigma + mu
        minmax_unscaled = array * max_val + min_val
        return np.where(min_val >= 0.0, minmax_unscaled, z_unscaled)

    def _extract_stats(self, df: pd.DataFrame, c_code: str, c_value: str) -> dict[str, pd.Series]:
        stat = df.groupby(c_code)[[c_value]].apply(
            lambda x: pd.Series(
                {"mu": x[c_value].mean(), "sigma": x[c_value].std(), "min": x[c_value].min(), "max": x[c_value].max()}
            )
        )
        return dict(
            mean=stat.loc[:, "mu"], std=stat.loc[:, "sigma"], min_val=stat.loc[:, "min"], max_val=stat.loc[:, "max"]
        )


class IQROutlierRemover(CodedValueProcessor):
    config: IQROutlierRemoverConfig
    min_val: pd.Series
    max_val: pd.Series

    def __init__(
        self,
        config: IQROutlierRemoverConfig,
        min_val: pd.Series = pd.Series(),
        max_val: pd.Series = pd.Series(),
        table_name: str | None = None,
        code_column: str | None = None,
        value_column: str | None = None,
    ):
        super().__init__(config=config, table_name=table_name, code_column=code_column, value_column=value_column)
        self.min_val = min_val
        self.max_val = max_val

    def __call__(self, dataset: Dataset) -> Dataset:
        table = self.table_getter(dataset)

        min_val = table.loc[:, self.code_column].map(self.min_val)
        max_val = table.loc[:, self.code_column].map(self.max_val)
        table = table.loc[table.loc[:, self.value_column].between(min_val, max_val)]

        return eqx.tree_at(self.table_getter, dataset, table)

    def _extract_stats(self, df: pd.DataFrame, c_code: str, c_value: str) -> dict[str, pd.Series]:
        outlier_q = np.array([self.config.outlier_q1, self.config.outlier_q2])
        q = df.groupby(c_code)[[c_value]].apply(lambda x: x[c_value].quantile(outlier_q))

        q.columns = ["q1", "q2"]
        q["iqr"] = q["q2"] - q["q1"]
        q["out_q1"] = q["q1"] - self.config.outlier_iqr_scale * q["iqr"]
        q["out_q2"] = q["q2"] + self.config.outlier_iqr_scale * q["iqr"]

        stat = df.groupby(c_code)[[c_value]].apply(
            lambda x: pd.Series({"mu": x[c_value].mean(), "sigma": x[c_value].std()})
        )

        stat["out_z1"] = stat["mu"] - self.config.outlier_z1 * stat["sigma"]
        stat["out_z2"] = stat["mu"] + self.config.outlier_z2 * stat["sigma"]
        return dict(min_val=np.minimum(q["out_q1"], stat["out_z1"]), max_val=np.maximum(q["out_q2"], stat["out_z2"]))  # type: ignore


class ObsIQROutlierRemover(TrainableTransformation):
    @classmethod
    def apply(
        cls, tvx_ehr: TVxEHR, schemes_context: CodingSchemesManager, report: TVxReport
    ) -> tuple[TVxEHR, TVxReport]:
        assert tvx_ehr.config.numerical_processors.outlier_removers is not None, (
            "tvx_ehr.config.numerical_processors.outlier_removers is None"
        )
        config = tvx_ehr.config.numerical_processors.outlier_removers.obs
        if config is None:
            return cls.skip(tvx_ehr, report, "config.numerical_processors.outlier_removers.obs is None")
        remover = IQROutlierRemover(config=config).fit(
            tvx_ehr.dataset,
            cls.get_admission_ids(tvx_ehr),
            table_name="obs",
            code_column=tvx_ehr.dataset.config.columns.obs.code,
            value_column=tvx_ehr.dataset.config.columns.obs.measurement,
        )
        tvx_ehr = eqx.tree_at(
            lambda x: x.numerical_processors.outlier_removers.obs, tvx_ehr, remover, is_leaf=lambda x: x is None
        )
        report = report.add(
            table="obs",
            column=None,
            value_type="type",
            transformation=cls,
            operation="TVxEHR.numerical_processors.outlier_removers.obs <- IQROutlierRemover",
            after=type(remover),
        )
        assert tvx_ehr.dataset.tables.obs is not None, "tvx_ehr.dataset.tables.obs is None"
        n1 = len(tvx_ehr.dataset.tables.obs)
        # TODO: report specific removals stats for each code.
        tvx_ehr = eqx.tree_at(lambda x: x.dataset, tvx_ehr, remover(tvx_ehr.dataset))
        assert tvx_ehr.dataset.tables.obs is not None, "tvx_ehr.dataset.tables.obs is None"
        n2 = len(tvx_ehr.dataset.tables.obs)
        report = report.add(
            table="obs", column=None, value_type="count", transformation=cls, operation="filter", before=n1, after=n2
        )
        return tvx_ehr, report


class ObsAdaptiveScaler(TrainableTransformation):
    @classmethod
    def apply(
        cls, tvx_ehr: TVxEHR, schemes_context: CodingSchemesManager, report: TVxReport
    ) -> tuple[TVxEHR, TVxReport]:
        assert tvx_ehr.config.numerical_processors.scalers is not None, (
            "tvx_ehr.config.numerical_processors.scalers is None"
        )
        config = tvx_ehr.config.numerical_processors.scalers.obs

        if config is None:
            return cls.skip(tvx_ehr, report, "config.numerical_processors.scalers.obs is None")

        value_column = tvx_ehr.dataset.config.columns.obs.measurement
        scaler = AdaptiveScaler(config=config).fit(
            tvx_ehr.dataset,
            cls.get_admission_ids(tvx_ehr),
            table_name="obs",
            code_column=tvx_ehr.dataset.config.columns.obs.code,
            value_column=value_column,
        )
        tvx_ehr = eqx.tree_at(
            lambda x: x.numerical_processors.scalers.obs, tvx_ehr, scaler, is_leaf=lambda x: x is None
        )
        report = report.add(
            table="obs",
            column=None,
            value_type="type",
            transformation=cls,
            operation="TVxEHR.numerical_processors.scalers.obs <- AdaptiveScaler",
            after=type(scaler),
        )
        assert tvx_ehr.dataset.tables.obs is not None, "tvx_ehr.dataset.tables.obs is None"
        dtype1 = tvx_ehr.dataset.tables.obs[value_column].dtype
        tvx_ehr = eqx.tree_at(lambda x: x.dataset, tvx_ehr, scaler(tvx_ehr.dataset))
        assert tvx_ehr.dataset.tables.obs is not None, "tvx_ehr.dataset.tables.obs is None"
        dtype2 = tvx_ehr.dataset.tables.obs[value_column].dtype
        report = report.add(
            table="obs",
            column=value_column,
            value_type="dtype",
            transformation=cls,
            operation=f"scaled_and_maybe_cast_{scaler.config.use_float16}",
            before=dtype1,
            after=dtype2,
        )
        return tvx_ehr, report


class InputScaler(TrainableTransformation):
    @classmethod
    def apply(
        cls, tvx_ehr: TVxEHR, schemes_context: CodingSchemesManager, report: TVxReport
    ) -> tuple[TVxEHR, TVxReport]:
        code_column = tvx_ehr.dataset.config.columns.icu_inputs.code
        value_column = tvx_ehr.dataset.config.columns.icu_inputs.derived_normalized_amount_per_hour
        assert tvx_ehr.config.numerical_processors.scalers is not None, (
            "tvx_ehr.config.numerical_processors.scalers is None"
        )
        config = tvx_ehr.config.numerical_processors.scalers.icu_inputs

        if config is None:
            return cls.skip(tvx_ehr, report, "config.numerical_processors.scalers.icu_inputs is None")

        scaler = MaxScaler(config=config).fit(
            tvx_ehr.dataset,
            cls.get_admission_ids(tvx_ehr),
            table_name="icu_inputs",
            code_column=code_column,
            value_column=value_column,
        )

        tvx_ehr = eqx.tree_at(
            lambda x: x.numerical_processors.scalers.icu_inputs, tvx_ehr, scaler, is_leaf=lambda x: x is None
        )
        report = report.add(
            table="icu_inputs",
            column=None,
            value_type="type",
            operation="TVxEHR.numerical_processors.scalers.icu_inputs <- MaxScaler",
            after=type(scaler),
        )
        assert tvx_ehr.dataset.tables.icu_inputs is not None, "tvx_ehr.dataset.tables.icu_inputs is None"
        dtype1 = tvx_ehr.dataset.tables.icu_inputs[value_column].dtype
        tvx_ehr = eqx.tree_at(lambda x: x.dataset, tvx_ehr, scaler(tvx_ehr.dataset))
        assert tvx_ehr.dataset.tables.icu_inputs is not None, "tvx_ehr.dataset.tables.icu_inputs is None"
        dtype2 = tvx_ehr.dataset.tables.icu_inputs[value_column].dtype
        report = report.add(
            table="icu_inputs",
            column=value_column,
            value_type="dtype",
            transformation=cls,
            operation=f"scaled_and_maybe_cast_{scaler.config.use_float16}",
            before=dtype1,
            after=dtype2,
        )
        return tvx_ehr, report


class InterventionSegmentation(TVxTransformation):
    @classmethod
    def apply(
        cls, tvx_ehr: TVxEHR, schemes_context: CodingSchemesManager, report: TVxReport
    ) -> tuple[SegmentedTVxEHR | TVxEHR, TVxReport]:
        if tvx_ehr.config.interventions_segmentation in (None, False) or not tvx_ehr.config.interventions:
            reason = []
            if tvx_ehr.config.interventions_segmentation in (None, False):
                reason.append("config.interventions_segmentation is None/False")
            if not tvx_ehr.config.interventions:
                reason.append("config.interventions is False")
            return cls.skip(tvx_ehr, report, reason=" + ".join(reason))

        maximum_padding = 100
        segmented_tvx_ehr = SegmentedTVxEHR.from_tvx_ehr(
            tvx_ehr, schemes_context=schemes_context, maximum_padding=maximum_padding
        )
        tvx_concept_path = TVxReportAttributes.admission_attribute_prefix("observables", InpatientObservables)
        report = report.add(
            tvx_concept=tvx_concept_path,
            value_type="concepts_count",
            operation=f"segmentation(maximum_padding={maximum_padding})",
            transformation=cls,
            before=sum(1 for _ in tvx_ehr.iter_obs()),
            after=sum(1 for _ in segmented_tvx_ehr.iter_obs()),
        )
        return segmented_tvx_ehr, report


class ObsTimeBinning(TVxTransformation):
    @classmethod
    def apply(
        cls, tvx_ehr: TVxEHR, schemes_context: CodingSchemesManager, report: TVxReport
    ) -> tuple[TVxEHR, TVxReport]:
        if tvx_ehr.config.time_binning is None:
            return cls.skip(tvx_ehr, report, reason="config.time_binning is None")

        interval = tvx_ehr.config.time_binning
        obs_scheme = tvx_ehr.scheme_proxy(schemes_context).obs
        assert isinstance(obs_scheme, NumericScheme), "obs_scheme is not NumericScheme"
        assert tvx_ehr.subjects is not None, "tvx_ehr.subjects is None"

        tvx_concept_path = TVxReportAttributes.admission_attribute_prefix("observables", InpatientObservables)
        tvx_binned_ehr = eqx.tree_at(
            lambda x: x.subjects,
            tvx_ehr,
            {
                subject_id: subject.observables_time_binning(interval, obs_scheme.types)
                for subject_id, subject in tvx_ehr.subjects.items()
            },
        )

        report = report.add(
            tvx_concept=tvx_concept_path,
            value_type="concepts_count",
            operation="time_binning",
            transformation=cls,
            before=sum(1 for _ in tvx_ehr.iter_obs()),
            after=sum(1 for _ in tvx_binned_ehr.iter_obs()),
        )
        report = report.add(
            tvx_concept=tvx_concept_path,
            value_type="timestamps_count",
            operation="time_binning",
            transformation=cls,
            before=sum(len(o) for o in tvx_ehr.iter_obs()),
            after=sum(len(o) for o in tvx_binned_ehr.iter_obs()),
        )
        report = report.add(
            tvx_concept=tvx_concept_path,
            transformation=cls,
            value_type="values_count",
            operation="time_binning",
            before=sum(o.count for o in tvx_ehr.iter_obs()),
            after=sum(o.count for o in tvx_binned_ehr.iter_obs()),
        )
        return tvx_binned_ehr, report


class LeadingObservableExtraction(TVxTransformation):
    @classmethod
    def apply(
        cls, tvx_ehr: TVxEHR, schemes_context: CodingSchemesManager, report: TVxReport
    ) -> tuple[TVxEHR, TVxReport]:
        if tvx_ehr.config.leading_observable is None:
            return cls.skip(tvx_ehr, report, reason="config.leading_observable is None")
        obs_scheme = tvx_ehr.dataset.scheme_proxy(schemes_context).obs
        assert isinstance(obs_scheme, NumericScheme), "obs_scheme is not NumericScheme"
        assert tvx_ehr.subjects is not None, "tvx_ehr.subjects is None"
        extractor = LeadingObservableExtractor(tvx_ehr.config.leading_observable, observable_scheme=obs_scheme)
        tvx_concept_path = TVxReportAttributes.admission_attribute_prefix("leading_observables", InpatientObservables)
        tvx_ehr = eqx.tree_at(
            lambda x: x.subjects,
            tvx_ehr,
            {
                subject_id: subject.extract_leading_observables(extractor)
                for subject_id, subject in tvx_ehr.subjects.items()
            },
        )

        report = report.add(
            tvx_concept=tvx_concept_path,
            value_type="concepts_count",
            operation="LeadingObservableExtractor",
            transformation=cls,
            after=sum(1 for _ in tvx_ehr.iter_lead_obs()),
        )
        report = report.add(
            tvx_concept=tvx_concept_path,
            value_type="timestamps_count",
            operation="LeadingObservableExtractor",
            transformation=cls,
            after=sum(len(lo) for lo in tvx_ehr.iter_lead_obs()),
        )
        report = report.add(
            tvx_concept=tvx_concept_path,
            transformation=cls,
            value_type="values_count",
            operation="LeadingObservableExtractor",
            after=sum(lo.count for lo in tvx_ehr.iter_lead_obs()),
        )
        report = report.add(
            transformation=cls,
            tvx_concept=tvx_concept_path,
            value_type="type",
            operation="LeadingObservableExtractor",
            after=InpatientObservables,
        )
        return tvx_ehr, report


class TVxConcepts(TVxTransformation):
    @classmethod
    def _static_info(
        cls, tvx_ehr: TVxEHR, schemes_context: CodingSchemesManager, report: TVxReport
    ) -> tuple[dict[str, StaticInfo], TVxReport]:
        tvx_scheme_proxy = tvx_ehr.scheme_proxy(schemes_context)
        static = tvx_ehr.dataset.tables.static
        static_config = tvx_ehr.dataset.config.columns.static
        config = tvx_ehr.config.demographic
        c_gender = static_config.gender
        c_date_of_birth = static_config.date_of_birth
        c_ethnicity = static_config.race

        report = report.add(
            transformation=cls,
            table="static",
            column=None,
            value_type="count",
            operation="extract_static_info",
            after=len(static),
        )

        gender, ethnicity, dob = {}, {}, {}
        if c_date_of_birth in static.columns and config.age:
            dob = static[c_date_of_birth].to_dict()
        if tvx_scheme_proxy.gender is not None and config.gender:
            gender_m = tvx_scheme_proxy.gender_mapper(tvx_ehr.dataset.config.scheme)
            assert gender_m is not None, "gender_m is None"
            target_scheme = tvx_scheme_proxy.gender
            gender_dict = {
                subject_id: gender_m.map_codeset({c}) for subject_id, c in static[c_gender].to_dict().items()
            }
            gender = {subject_id: target_scheme.codeset2vec(codes) for subject_id, codes in gender_dict.items()}

        if tvx_scheme_proxy.ethnicity is not None and config.ethnicity:
            ethnicity_m = tvx_scheme_proxy.ethnicity_mapper(tvx_ehr.dataset.config.scheme)
            target_scheme = tvx_scheme_proxy.ethnicity
            assert ethnicity_m is not None, "ethnicity_m is None"
            ethnicity_dict = {
                subject_id: ethnicity_m.map_codeset({c}) for subject_id, c in static[c_ethnicity].to_dict().items()
            }
            ethnicity = {subject_id: target_scheme.codeset2vec(codes) for subject_id, codes in ethnicity_dict.items()}

        static_info = {
            subject_id: StaticInfo(
                date_of_birth=dob.get(subject_id), ethnicity=ethnicity.get(subject_id), gender=gender.get(subject_id)
            )
            for subject_id in static.index
        }
        report = report.add(
            tvx_concept=TVxReportAttributes.static_info_prefix(),
            transformation=cls,
            column=None,
            value_type="count",
            operation="extract_static_info",
            after=len(static_info),
        )
        return static_info, report

    @staticmethod
    def _dx_discharge(
        tvx_ehr: TVxEHR, schemes_context: CodingSchemesManager
    ) -> tuple[dict[str, CodesVector], dict[str, frozenset[str]]]:
        tvx_scheme_proxy = tvx_ehr.scheme_proxy(schemes_context)
        c_adm_id = tvx_ehr.dataset.config.columns.dx_discharge.admission_id
        c_code = tvx_ehr.dataset.config.columns.dx_discharge.code
        df = tvx_ehr.dataset.tables.dx_discharge
        assert isinstance(df, pd.DataFrame), "tvx_ehr.dataset.tables.dx_discharge is not a DataFrame"
        dx_mapper = tvx_scheme_proxy.dx_mapper(tvx_ehr.dataset.config.scheme)
        assert dx_mapper is not None, "dx_mapper is None"
        target_scheme = tvx_scheme_proxy.dx_discharge
        assert target_scheme is not None, "target_scheme is None"
        n1 = len(df)
        dx_discharge = df[df[c_code].isin(dx_mapper.data)]
        n2 = len(dx_discharge)
        if n1 != n2:
            unique_codes_a = set(df[c_code])
            unique_codes_b = set(df[c_code]) & set(dx_mapper.data.keys())
            unique_removed = unique_codes_a - unique_codes_b
            n_uniq_a = len(unique_codes_a)
            n_uniq_rem = len(unique_removed)
            source_scheme = tvx_ehr.dataset.scheme_proxy(schemes_context).dx_discharge
            assert source_scheme is not None, "source_scheme is None"
            dataframe_log.info(
                f"In mapping ({dx_mapper.source_name}->{dx_mapper.target_name}), "
                f"some codes are not mapped to the target scheme.\n"
                f"{n1 - n2} / {n1} = {(n1 - n2) / n1: .2f} rows were removed.\n"
                f"{n_uniq_rem} / {n_uniq_a} = {n_uniq_rem / n_uniq_a: .2f} "
                f"unique codes were removed (see report).",
                dataframe=pd.DataFrame(
                    [(code, source_scheme.desc[code]) for code in unique_removed],
                    columns=pd.Series(["code", "description"]),
                ),
                tag="lost_dx_discharge_codes_unique",
            )

        dx_codes_set = dx_discharge.groupby(c_adm_id)[c_code].apply(set).to_dict()
        dx_codes_set = {k: dx_mapper.map_codeset(v) for k, v in dx_codes_set.items()}
        dx_codes_set = {adm_id: dx_codes_set.get(adm_id, frozenset()) for adm_id in tvx_ehr.admission_ids}

        return {adm_id: target_scheme.codeset2vec(codeset) for adm_id, codeset in dx_codes_set.items()}, dx_codes_set

    @staticmethod
    def _dx_discharge_history(
        tvx_ehr: TVxEHR, schemes_context: CodingSchemesManager, dx_discharge: dict[str, CodesVector]
    ) -> dict[str, CodesVector]:
        # TODO: test anti causality.
        dx_scheme = tvx_ehr.scheme_proxy(schemes_context).dx_discharge
        assert dx_scheme is not None, "tvx_ehr.scheme_proxy(...).dx_discharge is None"
        # For each subject accumulate previous dx_discharge codes.
        dx_discharge_history = dict()
        initial_history = dx_scheme.codeset2vec(set())
        # For each subject get the list of adm sorted by admission date.
        for subject_id, adm_ids in tvx_ehr.subjects_sorted_admission_ids.items():
            current_history = initial_history
            for adm_id in adm_ids:
                dx_discharge_history[adm_id] = current_history
                current_history = dx_discharge[adm_id].union(current_history)
        return dx_discharge_history

    @staticmethod
    def _outcome(
        tvx_ehr: TVxEHR, schemes_context: CodingSchemesManager, dx_discharge: dict[str, frozenset[str]]
    ) -> dict[str, CodesVector]:
        tvx_scheme_proxy = tvx_ehr.scheme_proxy(schemes_context)
        f_outcome = tvx_scheme_proxy.outcome
        assert f_outcome is not None, "tvx_scheme_proxy.outcome is None"
        return {str(adm_id): f_outcome(codeset) for adm_id, codeset in dx_discharge.items()}

    @staticmethod
    def _icu_inputs(tvx_ehr: TVxEHR, schemes_context: CodingSchemesManager) -> dict[str, InpatientInput]:
        table_config = tvx_ehr.dataset.config.columns.icu_inputs
        c_admission_id = table_config.admission_id
        c_code = table_config.code
        c_rate = table_config.derived_normalized_amount_per_hour
        c_start = table_config.start_time
        c_end = table_config.end_time

        # Here we avoid deep copy, and we can still replace
        # a new column without affecting the original table.
        assert tvx_ehr.dataset.tables.icu_inputs is not None, "tvx_ehr.dataset.tables.icu_inputs is None"
        table = tvx_ehr.dataset.tables.icu_inputs.iloc[:, :]
        icu_scheme = tvx_ehr.dataset.scheme_proxy(schemes_context).icu_inputs
        assert icu_scheme is not None, "tvx_ehr.scheme_proxy(...).icu_inputs is None"
        table[c_code] = table[c_code].map(icu_scheme.index)
        assert not table[c_code].isnull().any(), "Some codes are not in the target scheme."
        return {
            str(adm_id): InpatientInput(
                code_index=np.array(x[c_code].to_numpy(), dtype=np.int64),
                rate=x[c_rate].to_numpy(),
                starttime=x[c_start].to_numpy(),
                endtime=x[c_end].to_numpy(),
            )
            for adm_id, x in table.groupby(c_admission_id)
        }

    @staticmethod
    def _procedures(
        schemes_context: CodingSchemesManager,
        table: pd.DataFrame,
        config: AdmissionIntervalEventsTableColumns | AdmissionIntervalRatesTableColumns,
        code_map: CodeMap,
    ) -> dict[str, InpatientInput]:
        c_admission_id = config.admission_id
        c_code = config.code
        c_start_time = config.start_time
        c_end_time = config.end_time

        table = code_map.map_dataframe(table, c_code)
        target_index = code_map.target_index(schemes_context.scheme[code_map.target_name])
        table.loc[:, c_code] = table.loc[:, c_code].map(target_index)
        assert not table.loc[:, c_code].isnull().any(), "Some codes are not in the target scheme."

        return {
            str(adm_id): InpatientInput(
                code_index=np.array(x[c_code].to_numpy(), dtype=np.int64),
                rate=np.ones_like(x[c_code].to_numpy(), dtype=bool),
                starttime=x[c_start_time].to_numpy(),
                endtime=x[c_end_time].to_numpy(),
            )
            for adm_id, x in table.groupby(c_admission_id)
        }

    @staticmethod
    def _hosp_procedures(tvx_ehr: TVxEHR, schemes_context: CodingSchemesManager) -> dict[str, InpatientInput]:
        assert tvx_ehr.dataset.tables.hosp_procedures is not None, "tvx_ehr.dataset.tables.hosp_procedures is None"
        m = tvx_ehr.hosp_procedures_mapper(tvx_ehr.scheme_proxy(schemes_context))
        assert m is not None, "tvx_ehr.hosp_procedures_mapper(...) is None"
        return TVxConcepts._procedures(
            schemes_context,
            tvx_ehr.dataset.tables.hosp_procedures,
            tvx_ehr.dataset.config.columns.hosp_procedures,
            m,
        )

    @staticmethod
    def _icu_procedures(tvx_ehr: TVxEHR, schemes_context: CodingSchemesManager) -> dict[str, InpatientInput]:
        assert tvx_ehr.dataset.tables.icu_procedures is not None, "tvx_ehr.dataset.tables.icu_procedures is None"
        m = tvx_ehr.icu_procedures_mapper(tvx_ehr.scheme_proxy(schemes_context))
        assert m is not None, "tvx_ehr.icu_procedures_mapper(...) is None"
        return TVxConcepts._procedures(
            schemes_context,
            tvx_ehr.dataset.tables.icu_procedures,
            tvx_ehr.dataset.config.columns.icu_procedures,
            m,
        )

    @classmethod
    def _interventions(
        cls, tvx_ehr: TVxEHR, schemes_context: CodingSchemesManager, report: TVxReport
    ) -> tuple[dict[str, InpatientInterventions], TVxReport]:
        concept_path = TVxReportAttributes.inpatient_input_prefix
        hosp_procedures = cls._hosp_procedures(tvx_ehr, schemes_context)
        report = report.add(
            tvx_concept=concept_path("hosp_procedures"),
            transformation=cls,
            table="hosp_procedures",
            column=None,
            value_type="count",
            operation="extract_hosp_procedures",
            after=len(hosp_procedures),
        )

        icu_procedures = cls._icu_procedures(tvx_ehr, schemes_context)
        report = report.add(
            tvx_concept=concept_path("icu_procedures"),
            table="icu_procedures",
            transformation=cls,
            column=None,
            value_type="count",
            operation="extract_icu_procedures",
            after=len(icu_procedures),
        )

        icu_inputs = cls._icu_inputs(tvx_ehr, schemes_context)
        report = report.add(
            tvx_concept=concept_path("icu_inputs"),
            table="icu_inputs",
            transformation=cls,
            column=None,
            value_type="count",
            operation="extract_icu_inputs",
            after=len(icu_inputs),
        )
        interventions = {
            admission_id: InpatientInterventions(
                hosp_procedures=hosp_procedures.get(admission_id),
                icu_procedures=icu_procedures.get(admission_id),
                icu_inputs=icu_inputs.get(admission_id),
            )
            for admission_id in tvx_ehr.admission_ids
        }
        return interventions, report

    @classmethod
    def _observables(
        cls, tvx_ehr: TVxEHR, schemes_context: CodingSchemesManager, report: TVxReport
    ) -> tuple[dict[str, InpatientObservables], TVxReport]:
        c_admission_id = tvx_ehr.dataset.config.columns.obs.admission_id
        c_code = tvx_ehr.dataset.config.columns.obs.code
        c_value = tvx_ehr.dataset.config.columns.obs.measurement
        c_timestamp = tvx_ehr.dataset.config.columns.obs.time

        obs_scheme = tvx_ehr.scheme_proxy(schemes_context).obs
        table = tvx_ehr.dataset.tables.obs
        assert isinstance(obs_scheme, NumericScheme), "obs_scheme is not NumericScheme"
        assert table is not None, "tvx_ehr.dataset.tables.obs is None"
        table = table.assign(**{c_code: table.loc[:, c_code].map(obs_scheme.index)})
        assert not table.loc[:, c_code].isnull().any(), "Some codes are not in the target scheme."
        obs_dim = len(obs_scheme)
        tvx_concept_path = TVxReportAttributes.admission_attribute_prefix("observables", InpatientObservables)

        def time_values(index: np.ndarray, values: np.ndarray) -> np.ndarray:
            value = np.zeros(obs_dim, dtype=np.float16)
            value[index] = values
            return value

        def time_mask(index: np.ndarray) -> np.ndarray:
            mask = np.zeros(obs_dim, dtype=bool)
            mask[index] = True
            return mask

        def inpatient_obs_data(admission_df: pd.DataFrame) -> Generator[tuple[Hashable, np.ndarray, np.ndarray]]:
            admission_df = admission_df.sort_values(c_timestamp)
            for timestamp, time_df in admission_df.groupby(c_timestamp):
                index = time_df.loc[:, c_code].values
                yield timestamp, time_values(index, time_df[c_value].values), time_mask(index)  # type: ignore

        def make_inpatient_obs(admission_df: pd.DataFrame) -> InpatientObservables:
            time, value, mask = zip(*inpatient_obs_data(admission_df))
            return InpatientObservables(
                time=np.stack(time, axis=0),
                value=np.stack(value, dtype=np.float16, axis=0),
                mask=np.stack(mask, dtype=bool, axis=0),
            )

        inpatient_observables = table.groupby(c_admission_id)[[c_timestamp, c_code, c_value]].apply(make_inpatient_obs)  # type: ignore
        assert len(inpatient_observables.index.tolist()) == len(set(inpatient_observables.index.tolist())), (
            "Duplicate admission ids in obs"
        )
        inpatient_observables = inpatient_observables.to_dict()
        empty_obs = InpatientObservables.empty(
            size=obs_dim,
            time_dtype=next(iter(inpatient_observables.values())).time.dtype,
            value_dtype=np.float16,
            mask_dtype=bool,
        )
        empty_obs_dict = {adm_id: empty_obs for adm_id in tvx_ehr.admission_ids if adm_id not in inpatient_observables}

        report = report.add(table="obs", value_type="table_size", operation="extract_observables", after=len(table))
        report = report.add(
            tvx_concept=tvx_concept_path,
            table="obs",
            value_type="concepts_count",
            operation="extract_observables",
            transformation=cls,
            after=len(inpatient_observables),
        )
        report = report.add(
            tvx_concept=tvx_concept_path,
            table="obs",
            value_type="empty_concepts_count",
            operation="extract_observables",
            transformation=cls,
            after=len(empty_obs_dict),
        )
        report = report.add(
            tvx_concept=TVxReportAttributes.admission_attribute_prefix("observables", InpatientObservables),
            table="obs",
            value_type="timestamps_count",
            operation="extract_observables",
            transformation=cls,
            after=sum(len(o) for o in inpatient_observables.values()),
        )
        report = report.add(
            tvx_concept=TVxReportAttributes.admission_attribute_prefix("observables", InpatientObservables),
            transformation=cls,
            table="obs",
            value_type="values_count",
            operation="extract_observables",
            after=sum(o.count for o in inpatient_observables.values()),
        )

        return inpatient_observables | empty_obs_dict, report

    @classmethod
    def apply(
        cls, tvx_ehr: TVxEHR, schemes_context: CodingSchemesManager, report: TVxReport
    ) -> tuple[TVxEHR, TVxReport]:
        subject_admissions = tvx_ehr.subjects_sorted_admission_ids
        static_info, report = cls._static_info(tvx_ehr, schemes_context, report)

        dx_discharge, dx_discharge_codeset = cls._dx_discharge(tvx_ehr, schemes_context)
        dx_discharge_history = cls._dx_discharge_history(tvx_ehr, schemes_context, dx_discharge)
        outcome = cls._outcome(tvx_ehr, schemes_context, dx_discharge_codeset)
        if tvx_ehr.config.interventions:
            interventions, report = cls._interventions(tvx_ehr, schemes_context, report)
        else:
            interventions = None
        if tvx_ehr.config.observables:
            observables, report = cls._observables(tvx_ehr, schemes_context, report)
        else:
            observables = None

        def _admissions(admission_ids: list[str]) -> list[Admission]:
            return [
                Admission(
                    admission_id=i,
                    admission_dates=tvx_ehr.admission_dates[i],
                    dx_codes=dx_discharge[i],
                    dx_codes_history=dx_discharge_history[i],
                    outcome=outcome[i],
                    observables=observables[i] if observables else None,
                    interventions=interventions[i] if interventions else None,
                )
                for i in admission_ids
            ]

        subjects = {
            subject_id: Patient(
                subject_id=subject_id, admissions=_admissions(admission_ids), static_info=static_info[subject_id]
            )
            for subject_id, admission_ids in subject_admissions.items()
        }
        tvx_ehr = eqx.tree_at(lambda x: x.subjects, tvx_ehr, subjects, is_leaf=lambda x: x is None)
        report = report.add(
            tvx_concept=TVxReportAttributes.subjects_prefix(),
            transformation=cls,
            value_type="count",
            operation="extract_subjects",
            after=len(subjects),
        )
        return tvx_ehr, report



from typing import Literal

from ..coding_scheme import CodingSchemesManager
from ..dataset import AbstractDatasetPipeline, Dataset, DatasetColumns, DatasetConfig, DatasetSchemeConfig
from ..transformations import (
    CastTimestamps,
    FilterAdmissionsWithNoDiagnoses,
    FilterSubjectsNegativeAdmissionLengths,
    FilterSubjectsWithLongAdmission,
    FilterSubjectsWithSingleOrNoAdmission,
    FilterUnsupportedCodes,
    MergeOverlappingAdmissions,
    SetIndex,
    SqueezeToStandardColumns,
)
from ..tvx_concepts import DemographicVectorConfig
from ..tvx_ehr import AbstractTVxPipeline, TVxEHRConfig, TVxEHRSchemeConfig, TVxEHRSplitsConfig
from ..tvx_transformations import RandomSplits, SampleSubjects, TVxConcepts
from .mimic import load_mimic, MIMICDatasetAuxiliaryResources, MIMICDatasetSchemeSuffixes, ScopedSchemeNames
from .mimic_in_memory import (
    InMemoryMIMICTableFiles,
    MIMICIII_ADMISSIONS_COLMAP,
    MIMICIII_D_ICD_DIAGNOSES_COLMAP,
    MIMICIII_DIAGNOSES_ICD_COLMAP,
    MIMICIII_STATIC_COLMAP,
    MIMICIII_TABLES_RESOURCES,
    MIMICIV_ADMISSIONS_COLMAP,
    MIMICIV_D_ICD_DIAGNOSES_COLMAP,
    MIMICIV_DIAGNOSES_ICD_COLMAP,
    MIMICIV_STATIC_COLMAP,
    MIMICIV_TABLES_RESOURCES,
    MIMICTablesResources,
)


DatasetName = Literal["mimiciii", "mimiciv"]


def default_suffixes() -> MIMICDatasetSchemeSuffixes:
    return MIMICDatasetSchemeSuffixes(ethnicity="ethnicity", gender="gender", dx_discharge="dx_discharge")


def default_auxiliary_resources(dataset_name: DatasetName) -> MIMICTablesResources:
    return MIMICDatasetAuxiliaryResources.make_resources(
        name_prefix=f"{dataset_name}.dx_summary", resources_root="dx_summary_study", suffixes=default_suffixes()
    )


def default_dataset_schemes_config(scoped_names: ScopedSchemeNames) -> DatasetSchemeConfig:
    return DatasetSchemeConfig(
        ethnicity=scoped_names.ethnicity, gender=scoped_names.gender, dx_discharge=scoped_names.dx_discharge
    )


def default_dataset_config(scoped_names: ScopedSchemeNames) -> DatasetConfig:
    return DatasetConfig(
        scheme=default_dataset_schemes_config(scoped_names),
        columns=DatasetColumns(),
        select_subjects_with_observation=None,
        select_subjects_with_short_admissions=14.0,
    )  # two weeks.


def default_tvx_schemes_config(config: DatasetSchemeConfig) -> TVxEHRSchemeConfig:
    return TVxEHRSchemeConfig(gender=config.gender, ethnicity=None, dx_discharge="dx_ccs", outcome="dx_flat_ccs_v1")


def default_dataset_pipeline() -> AbstractDatasetPipeline:
    pipeline = [
        SqueezeToStandardColumns(),
        SetIndex(),
        CastTimestamps(),
        FilterSubjectsNegativeAdmissionLengths(),
        MergeOverlappingAdmissions(),
        FilterUnsupportedCodes(),
        FilterAdmissionsWithNoDiagnoses(),
        FilterSubjectsWithSingleOrNoAdmission(),
        FilterSubjectsWithLongAdmission(),
    ]
    return AbstractDatasetPipeline(transformations=pipeline)


def _mimic_from_memory(
    dataset_tables_resources: MIMICTablesResources,
    aux: MIMICDatasetAuxiliaryResources,
    schemes_config: DatasetSchemeConfig | None,
    in_memory_tables: InMemoryMIMICTableFiles,
) -> tuple[Dataset, CodingSchemesManager]:
    if schemes_config is None:
        schemes_config = default_dataset_schemes_config(aux.scoped_names)
    return load_mimic(
        config=DatasetConfig(scheme=schemes_config),
        tables=dataset_tables_resources,
        aux=aux,
        data_connection=in_memory_tables,
    )


def mimiciii_from_paths(
    patients: str,
    admissions: str,
    diagnoses_icd: str,
    d_icd_diagnoses: str,
    aux_resources: MIMICDatasetAuxiliaryResources = default_auxiliary_resources("mimiciii"),
    schemes_config: DatasetSchemeConfig | None = None,
    dataset_tables_resources: MIMICTablesResources = MIMICIII_TABLES_RESOURCES,
) -> tuple[Dataset, CodingSchemesManager]:
    # Optional argument. For faster loading time and lower memory footprint, only load the relevant raw columns.
    USECOLS = {
        "patients": tuple(MIMICIII_STATIC_COLMAP.keys()),
        "admissions": tuple(MIMICIII_ADMISSIONS_COLMAP.keys()),
        "diagnoses_icd": tuple(MIMICIII_DIAGNOSES_ICD_COLMAP.keys()),
        "d_icd_diagnoses": tuple(MIMICIII_D_ICD_DIAGNOSES_COLMAP.keys()),
    }
    in_memory_tables = InMemoryMIMICTableFiles.from_path(
        patients=patients,
        admissions=admissions,
        diagnoses_icd=diagnoses_icd,
        d_icd_diagnoses=d_icd_diagnoses,
        usecols=USECOLS,
    )
    return _mimic_from_memory(
        schemes_config=schemes_config,
        dataset_tables_resources=dataset_tables_resources,
        in_memory_tables=in_memory_tables,
        aux=aux_resources,
    )


def mimiciv_from_paths(
    patients: str,
    admissions: str,
    diagnoses_icd: str,
    d_icd_diagnoses: str,
    aux_resources: MIMICDatasetAuxiliaryResources = default_auxiliary_resources("mimiciv"),
    schemes_config: DatasetSchemeConfig | None = None,
    dataset_tables_resources: MIMICTablesResources = MIMICIV_TABLES_RESOURCES,
) -> tuple[Dataset, CodingSchemesManager]:
    # Optional argument. For faster loading time and lower memory footprint, only load the relevant raw columns.
    USECOLS = {
        "patients": tuple(MIMICIV_STATIC_COLMAP.keys()),
        "admissions": tuple(MIMICIV_ADMISSIONS_COLMAP.keys()),
        "diagnoses_icd": tuple(MIMICIV_DIAGNOSES_ICD_COLMAP.keys()),
        "d_icd_diagnoses": tuple(MIMICIV_D_ICD_DIAGNOSES_COLMAP.keys()),
    }
    in_memory_tables = InMemoryMIMICTableFiles.from_path(
        patients=patients,
        admissions=admissions,
        diagnoses_icd=diagnoses_icd,
        d_icd_diagnoses=d_icd_diagnoses,
        usecols=USECOLS,
    )
    return _mimic_from_memory(
        schemes_config=schemes_config,
        dataset_tables_resources=dataset_tables_resources,
        in_memory_tables=in_memory_tables,
        aux=aux_resources,
    )


def match_gender_schemes(
    scheme_config_a: DatasetSchemeConfig, scheme_config_b: DatasetSchemeConfig, schemes: CodingSchemesManager
) -> CodingSchemesManager:
    return schemes.add_match_map(scheme_config_a.gender, scheme_config_b.gender)


def default_tvx_ehr_config(config: DatasetSchemeConfig) -> TVxEHRConfig:
    scheme = default_tvx_schemes_config(config)
    return TVxEHRConfig(
        scheme=scheme,
        demographic=DemographicVectorConfig(age=True, gender=True, ethnicity=False),
        sample=None,  # no subsetting now
        splits=TVxEHRSplitsConfig(
            split_quantiles=[0.7, 0.85],
            seed=0,  # 0.7:0.15:0.15
            discount_first_admission=False,
            balance="admissions",
        ),
    )


def default_tvx_ehr_pipeline() -> AbstractTVxPipeline:
    pipeline = [SampleSubjects(), RandomSplits(), TVxConcepts()]
    return AbstractTVxPipeline(transformations=pipeline)



from . import (
    mimic as mimic,
    mimic_in_memory as mimic_in_memory,
    mimic_sql as mimic_sql,
    study_mimic_aki as study_mimic_aki,
    study_mimic_dx_summary as study_mimic_dx_summary,
)



#
#
# import pandas as pd
#
# from ..coding_scheme import CodingScheme
# from ..utils import resources_dir
#
# _CPRD_LTC_FILE = resources_dir('CPRD_212_LTC_ALL.csv.gz')
# ETH16_CODE_CNAME = 'eth16'
# ETH5_CODE_CNAME = 'eth5'
# ETH16_DESC_CNAME = 'eth16_desc'
# ETH5_DESC_CNAME = 'eth5_desc'
# ETH_SCHEME_FILE = 'cprd_eth.csv'
#
#
# class DxLTC212FlatCodes(CodingScheme):
#     _medcodes: dict[str, set[str]]
#     _system: dict[str, str]
#
#     def __init__(self, config: CodingSchemeConfig,
#                  codes: list[str], desc: dict[str, str], system: dict[str, str],
#                  medcodes: dict[str, set[str]]):
#         super().__init__(config=config,
#                          codes=codes,
#                          desc=desc)
#         self._system = system
#         self._medcodes = medcodes
#
#     @property
#     def medcodes(self):
#         return self._medcodes
#
#     @property
#     def system(self):
#         return self._system
#
#     @classmethod
#     def from_file(cls, name: str = 'dx_cprd_ltc212', filepath: str = _CPRD_LTC_FILE):
#         df = pd.read_csv(filepath, dtype=str)
#
#         medcode_cname = 'medcodeid'
#         disease_num_cname = 'disease_num'
#         disease_cname = 'disease'
#
#         system_cname = 'system'
#         system_num_cname = 'system_num'
#
#         desc = dict()
#         system = dict()
#         medcodes = dict()
#         for disease_num, disease_df in df.groupby(disease_num_cname):
#             disease_set = set(disease_df[disease_cname])
#             assert len(disease_set) == 1, "Disease name should be unique"
#             (disease_name,) = disease_set
#
#             system_set = set(disease_df[system_cname])
#             system_num_set = set(disease_df[system_num_cname])
#             assert len(system_set) == 1, "System name should be unique"
#             assert len(system_num_set) == 1, "System num should be unique"
#
#             (system_name,) = system_set
#             (system_num,) = system_num_set
#
#             medcodes_list = sorted(set(disease_df[medcode_cname]))
#
#             desc[disease_num] = disease_name
#             system[disease_num] = system_num
#             medcodes[disease_num] = medcodes_list
#
#         codes = sorted(set(df[disease_num_cname]))
#         return cls(config=CodingSchemeConfig(name),
#                    codes=codes,
#                    desc=desc,
#                    system=system,
#                    medcodes=medcodes)
#
#
# class DxLTC9809FlatMedcodes(CodingScheme):
#     _diseases: dict[str, list[str]]
#     _systems: dict[str, list[str]]
#     _diseases_desc: dict[str, str]
#     _systems_desc: dict[str, str]
#
#     def __init__(self, config: CodingSchemeConfig,
#                  codes: list[str], desc: dict[str, str], diseases: dict[str, list[str]],
#                  systems: dict[str, list[str]], diseases_desc: dict[str, str], systems_desc: dict[str, str]):
#         super().__init__(config=config,
#                          codes=codes,
#                          desc=desc)
#         self._diseases = diseases
#         self._systems = systems
#         self._diseases_desc = diseases_desc
#         self._systems_desc = systems_desc
#
#     @property
#     def systems(self):
#         return self._systems
#
#     @property
#     def diseases(self):
#         return self._diseases
#
#     @property
#     def diseases_desc(self):
#         return self._diseases_desc
#
#     @property
#     def systems_desc(self):
#         return self._systems_desc
#
#     @classmethod
#     def from_file(cls, name: str = 'dx_cprd_ltc9809', filepath: str = _CPRD_LTC_FILE):
#         filepath = resources_dir(filepath)
#         df = pd.read_csv(filepath, dtype=str)
#
#         medcode_cname = 'medcodeid'
#         disease_num_cname = 'disease_num'
#         disease_cname = 'disease'
#         desc_cname = 'descr'
#
#         system_cname = 'system'
#         system_num_cname = 'system_num'
#
#         desc = dict()
#         systems = dict()
#         diseases = dict()
#         for medcodeid, medcode_df in df.groupby(medcode_cname):
#             desc[medcodeid] = medcode_df[desc_cname].iloc[0]
#             systems[medcodeid] = sorted(set(medcode_df[system_num_cname]))
#             diseases[medcodeid] = sorted(set(medcode_df[disease_num_cname]))
#
#         codes = sorted(set(df[medcode_cname]))
#         diseases_desc = df.groupby(disease_num_cname)[desc_cname].apply(list).to_dict()
#         systems_desc = df.groupby(system_num_cname)[desc_cname].apply(list).to_dict()
#         return cls(config=CodingSchemeConfig(name),
#                    codes=codes,
#                    desc=desc,
#                    systems=systems,
#                    diseases=diseases,
#                    diseases_desc=diseases_desc,
#                    systems_desc=systems_desc)
#
#
# def register_cprd_ethnicity(name, eth_code_colname, eth_desc_colname):
#     filepath = resources_dir(ETH_SCHEME_FILE)
#     df = pd.read_csv(filepath, dtype=str)
#     desc = dict()
#     for eth_code, eth_df in df.groupby(eth_code_colname):
#         eth_set = set(eth_df[eth_desc_colname])
#         assert len(eth_set) == 1, "Ethnicity description should be unique"
#         (eth_desc,) = eth_set
#         desc[eth_code] = eth_desc
#
#     codes = sorted(set(df[eth_code_colname]))
#     Ethnicity.register_scheme(Ethnicity(CodingSchemeConfig(name),
#                                         codes=codes,
#                                         desc=desc))
#
#
# def register_cprd_ethnicity_scheme_loaders():
#     Ethnicity.register_scheme_loader('eth_cprd_16',
#                                      lambda: register_cprd_ethnicity('eth_cprd_16', ETH16_CODE_CNAME,
#                                      ETH16_DESC_CNAME))
#     Ethnicity.register_scheme_loader('eth_cprd_5',
#                                      lambda: register_cprd_ethnicity('eth_cprd_5', ETH5_CODE_CNAME, ETH5_DESC_CNAME))
#
#
# def register_cprd_eth_mapping():
#     filepath = resources_dir(ETH_SCHEME_FILE)
#     df = pd.read_csv(filepath, dtype=str)
#     data = df.set_index(ETH16_CODE_CNAME)[ETH5_CODE_CNAME].to_dict()
#     CodeMap.register_map('eth_cprd_16', 'eth_cprd_5', CodeMap(CodeMapConfig('eth_cprd_16', 'eth_cprd_5'), data=data))
#
#
# def register_cprd_gender():
#     codes = ['0', '1', '2']
#     desc = {
#         '0': 'female',
#         '1': 'male',
#         '2': 'intermediate'
#     }
#     SchemeWithMissing.register_scheme(
#         SchemeWithMissing(CodingSchemeConfig('cprd_gender'),
#                           codes=codes, desc=desc, missing_code='9'))
#
#
# def register_cprd_imd():
#     codes = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']
#     desc = dict(zip(codes, codes))
#     SchemeWithMissing.register_scheme(SchemeWithMissing(CodingSchemeConfig('cprd_imd_cat'),
#                                                         codes=codes, desc=desc,
#                                                         missing_code='99'))
#
#
# def register_medcode_mapping(medcode_scheme: str,
#                              disease_num_scheme: str):
#     medcode = CodingScheme.from_name(medcode_scheme)
#     CodeMap.register_map(medcode_scheme, disease_num_scheme,
#                          CodeMap(CodeMapConfig(medcode_scheme, disease_num_scheme), data=medcode.diseases))
#
#
# def setup_scheme_loaders():
#     CodingScheme.register_scheme_loader('dx_cprd_ltc212',
#                                         lambda: CodingScheme.register_scheme(DxLTC212FlatCodes.from_file()))
#     CodingScheme.register_scheme_loader('dx_cprd_ltc9809',
#                                         lambda: CodingScheme.register_scheme(DxLTC9809FlatMedcodes.from_file()))
#     register_cprd_ethnicity_scheme_loaders()
#     CodingScheme.register_scheme_loader('cprd_gender', register_cprd_gender)
#     CodingScheme.register_scheme_loader('cprd_imd_cat', register_cprd_imd)
#     for name in ('dx_cprd_ltc212_v1', 'dx_cprd_ltc9809_v1'):
#         FileBasedOutcomeExtractor.register_outcome_extractor_loader(name, f'{name}.json')
#
#
# def setup_maps_loaders():
#     # CPRD conversions
#     # LTC9809 -> LTC212
#     # Eth16 -> Eth5
#     CodeMap.register_map_loader('dx_cprd_ltc9809', 'dx_cprd_ltc212',
#                                 lambda: register_medcode_mapping('dx_cprd_ltc9809', 'dx_cprd_ltc212'))
#
#     CodeMap.register_map_loader('eth_cprd_16', 'eth_cprd_5', register_cprd_eth_mapping)
#
#
# def setup_cprd():
#     setup_scheme_loaders()
#     setup_maps_loaders()



from dataclasses import dataclass
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from ..coding_scheme import CodingSchemesManager


@dataclass
class CodingSchemeManagerStatisticsInterface:
    m: "CodingSchemesManager"

    def __init__(self, m: "CodingSchemesManager"):
        self.m = m



from collections.abc import Callable
from types import MappingProxyType
from typing import Any

import ehrax as rx
import equinox as eqx
import numpy as np
import pandas as pd
import pytest
import tables as tb
from ehrax.base import MAX_SEGMENT_SIZE
from tables import PerformanceWarning


# (A) Test ModuleMeta and AbstractModule
class Module(rx.AbstractModule):
    att1: Any
    att2: Any

    def __init__(self, att1: Any, att2: Any):
        self.att1 = att1
        self.att2 = att2


class TestAbstractModule:
    def test_module_registration(self):
        assert Module in rx.base._factory_registry.values()
        assert Module.__class_key__() == f"{self.__module__}.Module"
        assert f"{self.__module__}.Module" in rx.base._factory_registry.keys()
        assert rx.base._factory_registry[Module.__class_key__()] == Module


# (B) Test Config
class Config(rx.AbstractConfig):
    att1: Any
    att2: Any

    def __init__(self, att1: Any, att2: Any):
        self.att1 = att1
        self.att2 = att2


class TestAbstractConfig:
    def test_equality(self):
        assert Config(3, 5) == Config(3, 5)
        assert Config(-1, 2) != Config(0, 2)
        assert Config(Config(10, 5), {"k": 45}) == Config(Config(10, 5), dict(k=45))
        assert Config(Config(10, 5), {"k": 45}) != Config(Config(10, 5), dict(j=45))
        assert Config(Config(10, 5), {"k": 45}) != Config(Config(10, 0), dict(k=45))
        assert Config(Config(0, 0), Config(1, 1)) != Config(Config(0, 0), 5)

    def test_as_dict(self):
        assert Config(3, 5).as_dict() == {"att1": 3, "att2": 5}
        assert Config(-1, 2).as_dict() != {"att1": 0, "att2": 2}
        assert Config(Config(10, 5), {"k": 45}).as_dict() == {"att1": {"att1": 10, "att2": 5}, "att2": {"k": 45}}
        assert Config(Config(10, 5), {"k": 45}).as_dict() != {"att1": {"att1": 10, "att2": 5}, "att2": {"j": 45}}
        assert Config(Config(10, 5), {"k": 45}).as_dict() != {"att1": {"att1": 10, "att2": 0}, "att2": {"j": 45}}
        assert Config(Config(0, 0), Config(1, 1)).as_dict() != {"att1": {"att1": 0}, "att2": 5}


# (C) Test AbstractWithPandasEquivalent


class WithDataframeEquivalent(rx.AbstractWithDataframeEquivalent):
    a: dict[str, tuple[int, str, float]]

    def __init__(self, a: dict[str, tuple[int, str, float]]):
        self.a = a

    def to_dataframe(self):
        if len(self.a) == 0:
            return pd.DataFrame(columns=["c1", "c2", "c3"])

        return pd.DataFrame(
            data=list(self.a.values()),
            columns=["c1", "c2", "c3"],
            index=pd.Index(list(self.a.keys()), dtype="str", name="theindex"),
        ).astype({"c1": "int", "c2": "str", "c3": "float"})

    @classmethod
    def from_dataframe(cls, df: pd.DataFrame):
        vals = zip(df.c1, df.c2, df.c3)
        return cls(dict(zip(df.index, vals)))


class WithSeriesEquivalent(rx.AbstractWithSeriesEquivalent):
    a: dict[str, float]

    def __init__(self, a: Any):
        self.a = a

    def to_series(self):
        return pd.Series(
            list(self.a.values()),
            index=pd.Index(list(self.a.keys()), name="indexooo", dtype="str"),
            dtype=float,
            name="7lw",
        )

    @classmethod
    def from_series(cls, sr: pd.Series):
        return cls(sr.to_dict())


class TestAbstractWithPandasEquivalent:
    @pytest.fixture(
        scope="class",
        params=[
            {"pressure": 1.0, "temperature": 2.0, "humidity": 3.0},
            {"a": 0},
            {},
        ],
    )
    def a_proper_sr_eq(self, request):
        return WithSeriesEquivalent(request.param)

    @pytest.fixture(
        scope="class",
        params=[
            {"mondy": (0, "work", 1.0), "teuesday": (1, "cook", 2.0), "wednesday": (3, "study", 0.0)},
            {"sunday": (25, "run", 8.0)},
            {},
        ],
    )
    def a_proper_df_eq(self, request):
        return WithDataframeEquivalent(request.param)

    def test_dataframe_equivalence(self, a_proper_df_eq: WithDataframeEquivalent):
        assert WithDataframeEquivalent.from_dataframe(a_proper_df_eq.to_dataframe()) == a_proper_df_eq

    def test_series_equivalence(self, a_proper_sr_eq: WithSeriesEquivalent):
        to_sr = a_proper_sr_eq.to_series()
        from_sr = WithSeriesEquivalent.from_series(to_sr)
        assert from_sr == a_proper_sr_eq
        assert to_sr.to_dict() == from_sr.to_series().to_dict()

    @pytest.fixture
    def deserialized_df_eq(self, a_proper_df_eq: WithDataframeEquivalent, hf5_group_writer: tb.Group):
        a_proper_df_eq.to_hdf_group(hf5_group_writer)
        hf5 = hf5_group_writer._v_file
        filename = hf5.filename
        group_path = hf5_group_writer._v_pathname
        hf5.close()
        with tb.open_file(filename, mode="r") as hf5:
            return WithDataframeEquivalent.from_hdf_group(hf5.get_node(group_path))

    @pytest.fixture
    def deserialized_sr_eq(self, a_proper_sr_eq: WithSeriesEquivalent, hf5_group_writer: tb.Group):
        a_proper_sr_eq.to_hdf_group(hf5_group_writer)
        hf5 = hf5_group_writer._v_file
        filename = hf5.filename
        group_path = hf5_group_writer._v_pathname
        hf5.close()
        with tb.open_file(filename, mode="r") as hf5:
            return WithSeriesEquivalent.from_hdf_group(hf5.get_node(group_path))

    def test_hdf_df_serialization(
        self, a_proper_df_eq: WithDataframeEquivalent, deserialized_df_eq: WithDataframeEquivalent
    ):
        assert a_proper_df_eq == deserialized_df_eq
        assert a_proper_df_eq.to_dataframe().equals(deserialized_df_eq.to_dataframe())

    def test_hdf_series_serialization(
        self, a_proper_sr_eq: WithSeriesEquivalent, deserialized_sr_eq: WithSeriesEquivalent
    ):
        assert a_proper_sr_eq == deserialized_sr_eq
        assert a_proper_sr_eq.to_series().equals(deserialized_sr_eq.to_series())


# (D) Test  rx.AbstractVxData


class VxData(rx.AbstractVxData):
    a: Any
    b: Any
    c: Any

    def __init__(self, a: Any, b: Any, c: Any):
        self.a = a
        self.b = b
        self.c = c


class TestVxData:
    @pytest.fixture(scope="class", params=[])
    def vxdata_pair(self, request) -> tuple[VxData, VxData]:
        assert False, "Should be subclassed."

    @pytest.fixture
    def deserialized_pair(self, vxdata_pair: tuple[VxData, VxData], hf5_writer_file: tb.File) -> tuple[VxData, VxData]:
        a, b = vxdata_pair
        a.to_hdf_group(hf5_writer_file.create_group("/", "a"))
        b.to_hdf_group(hf5_writer_file.create_group("/", "b"))
        filename = hf5_writer_file.filename
        hf5_writer_file.close()
        with tb.open_file(filename, mode="r") as hf5_reader:
            _a = VxData.from_hdf_group(hf5_reader.get_node("/", "a"))
            _b = VxData.from_hdf_group(hf5_reader.get_node("/", "b"))
            return _a, _b

    def test_vxdata_equalities(self, vxdata_pair: tuple[VxData, VxData]):
        a, b = vxdata_pair
        assert not a.equals(b)

    def test_vxdata_serialization(self, vxdata_pair: tuple[VxData, VxData], deserialized_pair: tuple[VxData, VxData]):
        a, b = vxdata_pair
        a_, b_ = deserialized_pair
        assert a.equals(a_)
        assert b.equals(b_)
        assert not a_.equals(b_)


class TestVxDataWithPlainTypes(TestVxData):
    # (D.2) Test serialization/deserialization
    #       * with numpy
    #       * with pandas
    #       * with plain data types
    #       * with panda timestamp
    #       * with iterables
    #       * with maps
    #       * disk io
    @pytest.fixture(
        scope="class",
        params=[
            # Case 1
            [(None, None, None), (None, None, 0)],
            # Case 2
            [(1, 2, 3), (1, 2, 2)],
            # Case 3
            [(1.0, 2.0, 3.0), (1.0, 2.0, 3.01)],
            # Case 4
            [("a", "b", "c"), ("a", "b", "c_")],
            # Case 5
            [(True, False, True), (True, False, 1)],
            # Case 6
            [(1, 2.0, False), (1, 2.0, 0)],
            # Case 7
            [(1.0, "a", pd.Timestamp(5)), (1.0, "a", pd.Timestamp(4))],
        ],
    )
    def vxdata_pair(self, request) -> tuple[VxData, VxData]:
        return VxData(*request.param[0]), VxData(*request.param[1])


class TestVxDataWithCollectionTypes(TestVxData):
    @pytest.fixture(
        scope="class",
        params=[
            [((4, 3), ["aa", "bb"], {3.0, 5.0}), ((4, 3), ["aa", "bb"], {3.0, 5.1})],
            [({5: {}}, {"a": ()}, {"z": []}), ({5: []}, {"a": ()}, {"z": []})],
        ],
    )
    def vxdata_pair(self, request) -> tuple[VxData, VxData]:
        return VxData(*request.param[0]), VxData(*request.param[1])


class TestVxDataWithNumpy(TestVxData):
    @pytest.fixture(
        scope="class",
        params=[
            # Case 1: element values
            [(np.arange(10), np.array([]), np.array(-1)), (np.arange(10), np.array([]), np.array(0))],
            # Case 2: dtype
            [
                (np.arange(10), np.array([], dtype=float), np.array(0)),
                (np.arange(10), np.array([], dtype=int), np.array(0)),
            ],
            # Case 3: extra dim of 1
            [(np.arange(10), np.array([]), np.array([0])), (np.arange(10), np.array([]), np.array(0))],
            # Case 4: partly nan
            [
                (np.arange(10), np.array([float("nan"), 0.0]), np.array(0)),
                (np.arange(10), np.array([0.0, 0.0]), np.array(0)),
            ],
        ],
    )
    def vxdata_pair(self, request) -> tuple[VxData, VxData]:
        return VxData(*request.param[0]), VxData(*request.param[1])


class TestVxDataWithPandas(TestVxData):
    @pytest.fixture(
        scope="class",
        params=[
            # Case 1: element values in dataframe
            [
                (pd.DataFrame([1, 2]), pd.DataFrame(), pd.Series([32])),
                (pd.DataFrame([1, 3]), pd.DataFrame(), pd.Series([32])),
            ],
            # Case 2: element value in series
            [
                (pd.DataFrame([1, 2]), pd.DataFrame(), pd.Series([32])),
                (pd.DataFrame([1, 2]), pd.DataFrame(), pd.Series([34])),
            ],
            # Case 3: columns in empty dataframe
            [
                (pd.DataFrame([1, 2]), pd.DataFrame(columns=["x"]), pd.Series([32])),
                (pd.DataFrame([1, 2]), pd.DataFrame(), pd.Series([32])),
            ],
            # Case 4: name in empty series
            [
                (pd.DataFrame([1, 2]), pd.DataFrame(), pd.Series()),
                (pd.DataFrame([1, 2]), pd.DataFrame(), pd.Series(name="x")),
            ],
        ],
    )
    def vxdata_pair(self, request) -> tuple[VxData, VxData]:
        return VxData(*request.param[0]), VxData(*request.param[1])


class TestVxDataWithNesting(TestVxData):
    @pytest.fixture(
        scope="class",
        params=[
            # Case 1
            [
                (
                    {
                        "a": 1,
                        "b": 2,
                    },  # --------------------------------------_------------------(!)--- <- the only change is 5
                    {
                        "c": VxData(2.0, VxData(np.arange(100), pd.Timestamp(4), [pd.Series(np.arange(5))]), {34: "x"}),
                        "d": VxData(None, pd.Timestamp(10), [Config(Config(6, None), False)]),
                    },
                    {pd.Timestamp(0), pd.Timestamp(100)},
                ),
                (
                    {"a": 1, "b": 2},
                    # ---------------------------------------------------------(!)--- <- changed to 4
                    {
                        "c": VxData(2.0, VxData(np.arange(100), pd.Timestamp(4), [pd.Series(np.arange(4))]), {34: "x"}),
                        "d": VxData(None, pd.Timestamp(10), [Config(Config(6, None), False)]),
                    },
                    {pd.Timestamp(0), pd.Timestamp(100)},
                ),
            ],
            # Case 2
            # array, series, tuple[array, ...] ----------(!)-
            [
                (np.arange(29), pd.Series([4, 2]), (np.array([]), np.array([5.0]))),
                (np.arange(29), pd.Series([4, 2]), (np.array(5), np.array([5.0]))),
            ],
            # Case 3
            [
                (
                    pd.DataFrame(columns=["x", "y"]),
                    pd.DataFrame([5, 1]),
                    MappingProxyType(
                        # -------------------------------------------------------------------------------(!)
                        {
                            1: pd.Series(
                                [], pd.Index([], dtype="bool", name="indexooo"), dtype="bool", name="veryboolseries"
                            )
                        }
                    ),
                ),
                (
                    pd.DataFrame(columns=["x", "y"]),
                    pd.DataFrame([5, 1]),
                    MappingProxyType(
                        # --------------------------------------------------------------------------------(!)
                        {
                            1: pd.Series(
                                [], pd.Index([], dtype="bool", name="indexooo"), dtype="bool", name="--------------"
                            )
                        }
                    ),
                ),
            ],
        ],
    )
    def vxdata_pair(self, request) -> tuple[VxData, VxData]:
        return VxData(*request.param[0]), VxData(*request.param[1])


# Test lazy-loading...


class TestHDFVirtualNode:
    @pytest.fixture(
        scope="class",
        params=[
            ("x", "y", None),  # Cannot contain None
            (VxData(None, None, None), "y", "z"),  # Cannot contain non-plain types
            (rx.HDFVirtualNode("x", "y", "z"), "y", "z"),  # Cannot contain a virtual node
        ],
    )
    def invalid_args(self, request) -> tuple[Any, Any, Any]:
        return request.param

    def test_invalid_args(self, invalid_args):
        with pytest.raises(AssertionError):
            rx.HDFVirtualNode(*invalid_args)

    @pytest.fixture(scope="class")
    def dummy_vnode(self):
        return rx.HDFVirtualNode("x", "y", "z")

    def test_attribute_access(self, dummy_vnode):
        assert dummy_vnode.filename == "x"
        assert dummy_vnode.parent_path == "y"
        assert dummy_vnode.type_enum == "z"
        with pytest.raises(AttributeError):
            _ = dummy_vnode.foo

    def test_equality(self, dummy_vnode):
        with pytest.raises(ValueError, match="You are trying to test equality with a virtual unfetched node"):
            _ = dummy_vnode.equals(dummy_vnode)

    def test_to_hdf(self, dummy_vnode):
        with pytest.raises(ValueError, match="You are trying to serialize an unfetched node"):
            dummy_vnode.to_hdf_group(None)

    def test_from_hdf(self):
        with pytest.raises(ValueError, match="You are trying to deserialize"):
            _ = rx.HDFVirtualNode.from_hdf_group(None)


COMPLETE_VX_DATA = VxData(
    a={"a": 1, "b": VxData(None, 1, "x")},
    b={
        "c": VxData(2.0, VxData(np.arange(100), pd.Timestamp(4), [pd.Series(np.arange(4))]), {34: "x"}),
        "d": VxData(None, pd.Timestamp(10), [Config(Config(6, None), False)]),
    },
    c={pd.Timestamp(0), pd.Timestamp(100), "yyy"},
)

GETTER_NODE_PAIR = (
    # a pair of getter lambda, and the expected node from `complete_vx_data` to be returned.
    (lambda x: x.a, dict(a=1, b=VxData(None, 1, "x"))),
    (lambda x: x.b["c"].b, VxData(np.arange(100), pd.Timestamp(4), [pd.Series(np.arange(4))])),
    (lambda x: x.b["d"], VxData(None, pd.Timestamp(10), [Config(Config(6, None), False)])),
    (lambda x: x.c, {pd.Timestamp(0), pd.Timestamp(100), "yyy"}),
)


def _cmp(a: Any, b: Any):
    if hasattr(a, "equals"):
        return a.equals(b)
    else:
        return a == b


class TestLazyLoading:
    @pytest.fixture(scope="class")
    def complete_vx_data(self) -> VxData:
        return COMPLETE_VX_DATA

    @pytest.fixture(scope="class", params=GETTER_NODE_PAIR)
    def getter_node_pair(self, request, complete_vx_data: VxData) -> tuple[Callable[[VxData], Any], Any]:
        # ensure we are writing the right getters.
        getter, node = request.param
        assert _cmp(node, getter(complete_vx_data))
        return getter, node

    @pytest.fixture(scope="class")
    def pruned_vx_data(self, complete_vx_data: VxData, getter_node_pair: tuple[Callable[[VxData], Any], Any]) -> VxData:
        getter, _ = getter_node_pair
        return eqx.tree_at(getter, complete_vx_data, rx.HDFVirtualNode("x", "y", "z"))

    @pytest.fixture
    def hdf_serialized_vxdata(self, complete_vx_data: VxData, tmp_path_factory) -> str:
        filename = tmp_path_factory.mktemp("vxdata").joinpath("vxdata.h5")
        complete_vx_data.save(filename, complevel=0)
        loaded = VxData.load(filename)
        assert complete_vx_data.equals(loaded)
        return str(filename)

    @pytest.fixture
    def hdf_deserialized_deferred_vxdata(
        self, hdf_serialized_vxdata: str, getter_node_pair: tuple[Callable[[VxData], Any], Any]
    ) -> VxData:
        getter, _ = getter_node_pair
        return VxData.load(hdf_serialized_vxdata, defer=(getter,))

    @pytest.fixture
    def hdf_deserialized_deferred_vxdata2(self, hdf_serialized_vxdata: str) -> VxData:
        getters, _ = zip(*GETTER_NODE_PAIR)
        # No all four nodes deferred at once.
        return VxData.load(hdf_serialized_vxdata, defer=getters)

    @pytest.fixture
    def hdf_deserialized_fetched_at_vxdata(
        self, hdf_deserialized_deferred_vxdata: VxData, getter_node_pair: tuple[Callable[[VxData], Any], Any]
    ) -> VxData:
        getter, _ = getter_node_pair
        return rx.fetch_at(getter, hdf_deserialized_deferred_vxdata)

    @pytest.fixture
    def hdf_deserialized_one_level_fetched_at_vxdata(
        self, hdf_deserialized_deferred_vxdata: VxData, getter_node_pair: tuple[Callable[[VxData], Any], Any]
    ) -> VxData:
        getter, _ = getter_node_pair
        return rx.fetch_one_level_at(getter, hdf_deserialized_deferred_vxdata)

    @pytest.fixture
    def hdf_deserialized_fetched_at_vxdata2(self, hdf_deserialized_deferred_vxdata2: VxData) -> VxData:
        getters, _ = zip(*GETTER_NODE_PAIR)
        return rx.fetch_at(getters, hdf_deserialized_deferred_vxdata2)

    @pytest.fixture
    def hdf_deserialized_fetched_all_vxdata(self, hdf_deserialized_deferred_vxdata: VxData) -> VxData:
        return rx.fetch_all(hdf_deserialized_deferred_vxdata)

    @pytest.fixture
    def hdf_deserialized_fetched_all_vxdata2(self, hdf_deserialized_deferred_vxdata2: VxData) -> VxData:
        return rx.fetch_all(hdf_deserialized_deferred_vxdata2)

    def test_invalid_to_hdf(self, pruned_vx_data: VxData, hf5_group_writer: tb.Group):
        with pytest.raises(ValueError, match="You are trying to serialize an unfetched node"):
            pruned_vx_data.to_hdf_group(hf5_group_writer)

    def test_invalid_equality(self, pruned_vx_data: VxData, complete_vx_data: VxData):
        assert not pruned_vx_data.equals(complete_vx_data)
        assert not complete_vx_data.equals(pruned_vx_data)
        with pytest.raises(ValueError, match="You are trying to test equality with a virtual unfetched node"):
            _ = pruned_vx_data.equals(pruned_vx_data)

    def test_defer(
        self, hdf_deserialized_deferred_vxdata: VxData, getter_node_pair: tuple[Callable[[VxData], Any], Any]
    ):
        getter, _ = getter_node_pair
        assert isinstance(getter(hdf_deserialized_deferred_vxdata), rx.HDFVirtualNode)

    def _aux_test_fetch(
        self,
        hdf_deserialized_fetched_vxdata: VxData,
        getter_node_pair: tuple[Callable[[VxData], Any], Any],
        complete_vx_data: VxData,
    ):
        getter, node = getter_node_pair
        assert _cmp(getter(hdf_deserialized_fetched_vxdata), node)
        assert _cmp(getter(complete_vx_data), getter(hdf_deserialized_fetched_vxdata))
        assert complete_vx_data.equals(hdf_deserialized_fetched_vxdata)

    def test_fetch_at(
        self,
        hdf_deserialized_fetched_at_vxdata: VxData,
        getter_node_pair: tuple[Callable[[VxData], Any], Any],
        complete_vx_data: VxData,
    ):
        self._aux_test_fetch(hdf_deserialized_fetched_at_vxdata, getter_node_pair, complete_vx_data)

    def test_fetch_all(
        self,
        hdf_deserialized_fetched_all_vxdata: VxData,
        getter_node_pair: tuple[Callable[[VxData], Any], Any],
        complete_vx_data: VxData,
    ):
        self._aux_test_fetch(hdf_deserialized_fetched_all_vxdata, getter_node_pair, complete_vx_data)

    def test_fetch_one_level_at(
        self,
        complete_vx_data: VxData,
        hdf_deserialized_one_level_fetched_at_vxdata: VxData,
        getter_node_pair: tuple[Callable[[VxData], Any], Any],
    ):
        getter, node = getter_node_pair
        # Type is equal but contents are not
        assert type(getter(hdf_deserialized_one_level_fetched_at_vxdata)) is type(node)
        assert not _cmp(getter(hdf_deserialized_one_level_fetched_at_vxdata), node)
        # Children themselves must be virtual nodes.

        # if the virtual node represents a sequence containing plain types, then it will be loaded!
        get_immediate_leaves = (
            (lambda x: eqx.tree_flatten_one_level(x)[0]) if type(node) is not set else (lambda x: list(x))
        )
        if set(map(type, get_immediate_leaves(node))).issubset(rx.base.SERIES_GROUPED_ELEMENT_TYPES):
            assert all(
                isinstance(child, rx.HDFVirtualNode)
                for child in get_immediate_leaves(getter(hdf_deserialized_one_level_fetched_at_vxdata))
            )
            assert hdf_deserialized_one_level_fetched_at_vxdata.equals(complete_vx_data)
        else:
            assert not hdf_deserialized_one_level_fetched_at_vxdata.equals(complete_vx_data)

    def test_fetch_all_after_fetch_one_level_at(
        self, complete_vx_data: VxData, hdf_deserialized_one_level_fetched_at_vxdata: VxData
    ):
        all_fetched = rx.fetch_all(hdf_deserialized_one_level_fetched_at_vxdata)
        assert all_fetched.equals(complete_vx_data)

    def test_fetch_at2(
        self,
        hdf_deserialized_fetched_at_vxdata2: VxData,
        hdf_deserialized_fetched_all_vxdata2: VxData,
        complete_vx_data: VxData,
    ):
        assert hdf_deserialized_fetched_at_vxdata2.equals(complete_vx_data)
        assert hdf_deserialized_fetched_all_vxdata2.equals(complete_vx_data)
        assert hdf_deserialized_fetched_at_vxdata2.equals(hdf_deserialized_fetched_all_vxdata2)


class TestLargeDataSegmentation:
    BAD_PERFORMANCE_N_ATTRS = 5000
    BAD_PERFORMANCE_N_CHILDREN = 17000

    @pytest.fixture(
        scope="class",
        params=[
            MAX_SEGMENT_SIZE - 1,
            MAX_SEGMENT_SIZE,
            MAX_SEGMENT_SIZE + 1,
            max(BAD_PERFORMANCE_N_ATTRS, BAD_PERFORMANCE_N_CHILDREN),
        ],
    )
    def wide_data(self, request):
        n = request.param
        return VxData({5: [VxData(0, {}, [])] * n}, {"a": ()}, {i: VxData(0, 12, i) for i in range(n)})

    def test_ensure_correctness_of_attributes_warning_capture(self, hf5_group_writer: tb.Group):
        with pytest.warns(PerformanceWarning, match="maximum number of attributes"):
            for i in range(self.BAD_PERFORMANCE_N_ATTRS):
                hf5_group_writer._v_attrs[f"x{i}"] = i

    def test_ensure_correctness_of_children_warning_capture(self, hf5_writer_file: tb.File):
        with pytest.warns(PerformanceWarning, match="maximum number of children"):
            for i in range(self.BAD_PERFORMANCE_N_CHILDREN):
                hf5_writer_file.create_group(f"/x{i}/y", "z", createparents=True)

    @pytest.mark.filterwarnings("error:.*maximum number of.*")  # children or attributes
    def test_no_serialization_performance_warning(self, wide_data: VxData, tmp_path_factory):
        filename = tmp_path_factory.mktemp("wide").joinpath("data.h5")
        wide_data.save(filename, complevel=0)
        loaded = VxData.load(filename)
        assert wide_data.equals(loaded)

    def test_lazy_loading(self, wide_data: VxData, tmp_path_factory):
        filename = tmp_path_factory.mktemp("wide").joinpath("data.h5")
        wide_data.save(filename, complevel=0)
        # node
        loaded = VxData.load(filename, defer=(lambda x: x.c,), levels=(1,))
        assert wide_data.equals(rx.fetch_all(loaded))



from collections.abc import Iterator

import ehrax as rx
import equinox as eqx
import numpy as np
import pandas as pd
import pytest
import tables as tb
from ehrax.testing.common_setup import (
    _admission,
    _admissions,
    _dataset_tables,
    _dx_codes,
    _dx_codes_history,
    _icu_inputs,
    _inpatient_interventions,
    _inpatient_observables,
    _outcome,
    _proc,
    _segmented_inpatient_interventions,
    _singular_codevec,
    _static_info,
    DATASET_CONFIG,
    DATASET_SCHEME_CONF,
    DATASET_SCHEME_MANAGER,
    leading_observables_extractor,
    OUTCOME_DATA,
    SCHEMES,
    TARGET_SCHEMES,
    TVXEHR_CONF,
)

from .common_setup import (
    INTERVENTIONS_MAX_N_ITEMS,
    OBS_MAX_N_TIMESTAMPS,
    ADMISSION_CONCEPT_MAX_STAY_DAYS,
    ADMISSION_CONCEPT_MAX_STAY_HOURS,
    ADMISSION_TABLES_MAX_STAY_DAYS,
)


class Dataset(rx.Dataset):
    @classmethod
    def load_tables(cls, config: rx.DatasetConfig, scheme: rx.DatasetSchemeProxy) -> rx.DatasetTables:
        raise NotImplementedError("Still not implemented.")

    @classmethod
    def make_default_pipeline(cls) -> rx.AbstractDatasetPipeline:
        raise NotImplementedError("No.")


@pytest.fixture(
    params=[(1, 0, 0), (1, 2, 0)],
    ids=lambda x: f"_{x[0]}_subjects_{x[0] * x[1]}_admissions_{x[0] * x[1] * x[2]}_records",
    scope="session",
)
def dataset_tables_without_records(request) -> rx.DatasetTables:
    return _dataset_tables(
        DATASET_SCHEME_CONF, DATASET_SCHEME_MANAGER, request.param, max_stay_days=ADMISSION_TABLES_MAX_STAY_DAYS
    )


@pytest.fixture(scope="session")
def dataset_without_records(dataset_tables_without_records: rx.DatasetTables) -> Dataset:
    return Dataset(tables=dataset_tables_without_records, config=DATASET_CONFIG)


@pytest.fixture(
    params=[(1, 2, 10), (300, 3, 25)],
    ids=lambda x: f"_{x[0]}_subjects_{x[0] * x[1]}_admissions_{x[0] * x[1] * x[2]}_records",
    scope="session",
)
def dataset_tables_with_records(request) -> rx.DatasetTables:
    return _dataset_tables(
        DATASET_SCHEME_CONF, DATASET_SCHEME_MANAGER, request.param, max_stay_days=ADMISSION_TABLES_MAX_STAY_DAYS
    )


@pytest.fixture(scope="session")
def dataset_with_records(dataset_tables_with_records: rx.DatasetTables) -> Dataset:
    return Dataset(tables=dataset_tables_with_records, config=DATASET_CONFIG)


@pytest.fixture(scope="session")
def large_dataset_tables():
    return _dataset_tables(
        DATASET_SCHEME_CONF, DATASET_SCHEME_MANAGER, (1000, 5, 1), max_stay_days=ADMISSION_TABLES_MAX_STAY_DAYS
    )


@pytest.fixture(scope="session")
def large_dataset(large_dataset_tables: rx.DatasetTables) -> Dataset:
    return Dataset(tables=large_dataset_tables, config=DATASET_CONFIG)


@pytest.fixture(scope="session")
def unit_converter_table(dataset_tables_with_records: rx.DatasetTables) -> pd.DataFrame | None:
    assert "icu_inputs" in dataset_tables_with_records.tables_dict or len(dataset_tables_with_records.icu_inputs) == 0
    c_code = str(rx.COLUMN.code)
    c_amount_unit = str(rx.COLUMN.amount_unit)
    c_norm_factor = str(rx.COLUMN.derived_unit_normalization_factor)
    c_universal_unit = str(rx.COLUMN.derived_universal_unit)
    icu_inputs = dataset_tables_with_records.icu_inputs

    table = pd.DataFrame(
        columns=[c_code, c_amount_unit],
        data=[(code, unit) for code, unit in icu_inputs.groupby([c_code, c_amount_unit]).groups.keys()],
    )
    table[c_norm_factor] = 1.0
    table[c_universal_unit] = table[c_amount_unit].map(
        {c: np.random.choice(df[c_amount_unit].unique(), size=1)[0] for c, df in table.groupby(c_code)}
    )
    table[c_norm_factor] = table[c_norm_factor].where(
        table[c_universal_unit] == table[c_amount_unit], np.random.choice([1e-3, 100, 10, 1e3])
    )
    return table


@pytest.fixture(scope="session")
def mimiciv_dataset_scheme_config() -> rx.DatasetSchemeConfig:
    return rx.DatasetSchemeConfig(
        ethnicity=SCHEMES["ethnicity"].name,
        gender=SCHEMES["gender"].name,
        dx_discharge=SCHEMES["dx_discharge"].name,
        icu_procedures=SCHEMES["icu_procedures"].name,
        icu_inputs=SCHEMES["icu_inputs"].name,
        obs=SCHEMES["obs"].name,
        hosp_procedures=SCHEMES["hosp_procedures"].name,
    )


@pytest.fixture(scope="session")
def mimiciv_dataset_config(mimiciv_dataset_scheme_config: rx.DatasetSchemeConfig) -> rx.DatasetConfig:
    return rx.DatasetConfig(scheme=mimiciv_dataset_scheme_config)


@pytest.fixture(scope="session")
def mimiciv_dataset_without_records(mimiciv_dataset_config, dataset_tables_without_records) -> Dataset:
    ds = Dataset(tables=dataset_tables_without_records, config=mimiciv_dataset_config)
    return eqx.tree_at(
        lambda x: x.columns, ds, dataset_tables_without_records, is_leaf=lambda x: x is None
    )._execute_pipeline(
        [rx.SetIndex(), rx.SynchronizeSubjects(), rx.CastTimestamps(), rx.SetAdmissionRelativeTimes()],
        DATASET_SCHEME_MANAGER,
    )


@pytest.fixture(scope="session")
def mimiciv_dataset(dataset_tables_with_records: rx.DatasetTables, unit_converter_table: pd.DataFrame) -> Dataset:
    config = eqx.tree_at(
        lambda x: x.scheme, DATASET_CONFIG, rx.DatasetSchemeConfig(**DATASET_CONFIG.scheme.scheme_fields())
    )
    ds = Dataset(tables=dataset_tables_with_records, config=config)
    return ds._execute_pipeline(
        [
            rx.SetIndex(),
            rx.SynchronizeSubjects(),
            rx.CastTimestamps(),
            rx.ICUInputRateUnitConversion(),
            rx.SetAdmissionRelativeTimes(),
        ],
        DATASET_SCHEME_MANAGER,
    )


@pytest.fixture(scope="session")
def tvx_ehr(mimiciv_dataset: Dataset) -> rx.TVxEHR:
    return rx.TVxEHR(dataset=mimiciv_dataset, config=TVXEHR_CONF)


@pytest.fixture(scope="session")
def tvx_ehr_without_records(mimiciv_dataset_without_records: Dataset) -> rx.TVxEHR:
    return rx.TVxEHR(dataset=mimiciv_dataset_without_records, config=TVXEHR_CONF)


@pytest.fixture
def hf5_writer_file(tmpdir) -> Iterator[tb.File]:
    # No compression for faster tests.
    with tb.open_file(tmpdir.join("test.h5"), "w", filters=tb.Filters(complevel=0)) as h5f:
        yield h5f


@pytest.fixture
def hf5_group_writer(hf5_writer_file: tb.File) -> tb.Group:
    return hf5_writer_file.create_group("/", "test")


@pytest.fixture(scope="session")
def gender() -> rx.CodesVector:
    return _singular_codevec(SCHEMES["gender"])


@pytest.fixture(scope="session")
def ethnicity() -> rx.CodesVector:
    return _singular_codevec(SCHEMES["ethnicity"])


@pytest.fixture(scope="session")
def static_info(ethnicity: rx.CodesVector, gender: rx.CodesVector) -> rx.StaticInfo:
    return _static_info(ethnicity, gender)


@pytest.fixture(scope="session")
def dx_codes():
    return _dx_codes(TARGET_SCHEMES["dx_discharge"])


@pytest.fixture(scope="session")
def dx_codes_history(dx_codes: rx.CodesVector):
    return _dx_codes_history(dx_codes)


@pytest.fixture(scope="session")
def outcome(dx_codes: rx.CodesVector):
    return _outcome(OUTCOME_DATA, DATASET_SCHEME_MANAGER, dx_codes)


@pytest.fixture(params=[0, 1, 301], scope="session", ids=["0-obs", "1-obs", "301-obs"])
def inpatient_observables(request):
    n_timestamps = request.param
    return _inpatient_observables(TARGET_SCHEMES["obs"], n_timestamps, los_hours=ADMISSION_CONCEPT_MAX_STAY_HOURS)


@pytest.fixture(params=[0, 1, 5], scope="session", ids=["0icuin", "1icuin", "5icuin"])
def icu_inputs(request) -> rx.InpatientInput:
    return _icu_inputs(SCHEMES["icu_inputs"], request.param, los_hours=ADMISSION_CONCEPT_MAX_STAY_HOURS)


@pytest.fixture(params=[0, 5], scope="session", ids=["0icuproc", "5icuproc"])
def icu_proc(request):
    return _proc(SCHEMES["icu_procedures"], request.param, los_hours=ADMISSION_CONCEPT_MAX_STAY_HOURS)


@pytest.fixture(params=[0, 5], scope="session", ids=["0hosproc", "5hosproc"])
def hosp_proc(request):
    return _proc(SCHEMES["hosp_procedures"], n_timestamps=request.param, los_hours=ADMISSION_CONCEPT_MAX_STAY_HOURS)


@pytest.fixture(params=[0, 1, 2, -1], scope="session")
def inpatient_interventions_with_a_none(hosp_proc, icu_proc, icu_inputs, request):
    whoisnull = request.param
    return _inpatient_interventions(
        None if whoisnull == 0 else hosp_proc,
        None if whoisnull == 1 else icu_proc,
        None if whoisnull == 2 else icu_inputs,
    )


@pytest.fixture(scope="session")
def inpatient_interventions(hosp_proc, icu_proc, icu_inputs):
    return _inpatient_interventions(hosp_proc, icu_proc, icu_inputs)


@pytest.fixture(scope="session")
def segmented_inpatient_interventions(
    inpatient_interventions_with_a_none: rx.InpatientInterventions,
) -> rx.SegmentedInpatientInterventions:
    return _segmented_inpatient_interventions(
        inpatient_interventions_with_a_none,
        hosp_proc_scheme=SCHEMES["hosp_procedures"],
        icu_proc_scheme=SCHEMES["icu_procedures"],
        icu_inputs_scheme=SCHEMES["icu_inputs"],
        maximum_padding=1,
        max_los_hours=ADMISSION_CONCEPT_MAX_STAY_HOURS,
    )


@pytest.fixture(scope="session")
def leading_observable(inpatient_observables: rx.InpatientObservables) -> rx.InpatientObservables:
    return leading_observables_extractor(observation_scheme=SCHEMES["obs"])(inpatient_observables)


@pytest.fixture(scope="session")
def admission(
    dx_codes: rx.CodesVector,
    dx_codes_history: rx.CodesVector,
    outcome: rx.CodesVector,
    inpatient_observables: rx.InpatientObservables,
    inpatient_interventions: rx.InpatientInterventions,
    leading_observable: rx.InpatientObservables,
) -> rx.Admission:
    admission_id = "test"
    return _admission(
        admission_id=admission_id,
        admission_date=pd.to_datetime("now"),
        dx_codes=dx_codes,
        dx_codes_history=dx_codes_history,
        outcome=outcome,
        observables=inpatient_observables,
        interventions=inpatient_interventions,
        leading_observable=leading_observable,
        los_days=ADMISSION_CONCEPT_MAX_STAY_DAYS,
    )


@pytest.fixture(scope="session")
def segmented_admission(admission: rx.Admission) -> rx.SegmentedAdmission:
    return rx.SegmentedAdmission.from_admission(
        admission=admission,
        maximum_padding=1,
        icu_inputs_size=len(SCHEMES["icu_inputs"]),
        icu_procedures_size=len(SCHEMES["icu_procedures"]),
        hosp_procedures_size=len(SCHEMES["hosp_procedures"]),
    )


@pytest.fixture(scope="session")
def segmented_patient(patient: rx.Patient) -> rx.SegmentedPatient:
    return rx.SegmentedPatient.from_patient(
        patient=patient,
        maximum_padding=1,
        icu_inputs_size=len(SCHEMES["icu_inputs"]),
        icu_procedures_size=len(SCHEMES["icu_procedures"]),
        hosp_procedures_size=len(SCHEMES["hosp_procedures"]),
    )


@pytest.fixture(params=[0, 10], scope="session", ids=["0adms", "10adms"])
def patient(request, static_info: rx.StaticInfo) -> rx.Patient:
    admissions = _admissions(
        n_admissions=request.param,
        dx_scheme=SCHEMES["dx_discharge"],
        outcome_extractor_=OUTCOME_DATA,
        observation_scheme=SCHEMES["obs"],
        icu_inputs_scheme=SCHEMES["icu_inputs"],
        icu_proc_scheme=SCHEMES["icu_procedures"],
        hosp_proc_scheme=SCHEMES["hosp_procedures"],
        dataset_scheme_manager=DATASET_SCHEME_MANAGER,
        max_los_days=ADMISSION_CONCEPT_MAX_STAY_DAYS,
        max_n_timestamps_obs=OBS_MAX_N_TIMESTAMPS,
        max_n_inputs=INTERVENTIONS_MAX_N_ITEMS,
    )
    return rx.Patient(subject_id="test", admissions=admissions, static_info=static_info)
